/**
 * \file gre/uiviewtool.h
 * \brief GRE VIEWTOOL class definition
 *
 * \if NODOC
 * $Id: uiviewtool.h_v 1.13 2011/11/28 17:22:44 mju Exp $
 *
 * $Log: uiviewtool.h_v $
 * Revision 1.13  2011/11/28 17:22:44  mju
 * Fix dox dependencies.
 *
 * Revision 1.12  2010/12/02 16:17:41  mju
 * Add activateByHotKey.
 *
 * Revision 1.11  2010/12/02 16:00:47  mju
 * Add hotKeyCode.
 *
 * Revision 1.10  2008/02/14 18:30:31  mju
 * Add isSuspended.
 *
 * Revision 1.9  2008/02/13 16:10:49  mju
 * Add mgr::findByName.
 *
 * Revision 1.8  2008/02/07 16:01:43  mju
 * Add deleteAllScripts and removeUI.
 *
 * Revision 1.7  2008/01/09 18:57:09  mju
 * Change isActive to not be inline.
 *
 * Revision 1.6  2007/12/17 15:05:24  mju
 * Add viewtooo::getDlgParent and getViewable.
 *
 * Revision 1.5  2007/12/10 18:51:45  mju
 * Add viewtool::isActive.
 *
 * Revision 1.4  2007/12/03 15:22:31  mju
 * Make viewtool friend of toolmgr.
 *
 * Revision 1.3  2007/11/29 22:32:11  mju
 * Don't use pointer for menu as need ref-counted copy.
 *
 * Revision 1.2  2007/10/11 22:35:23  mju
 * Add toolmgr.suspend/resume.
 *
 * Revision 1.1  2007/10/08 20:36:24  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_GRE_UIVIEWTOOL_H
#define	INC_GRE_UIVIEWTOOL_H

#ifndef	INC_GRE_UIVIEWCANVAS_H
#include <gre/uiviewcanvas.h>
#endif

#ifndef	INC_MGUI_MENU_H
#include <mgui/menu.h>
#endif

namespace GRE {
namespace UI {

class VIEWTOOL;

//===================================================================================================================

//! Manager for tools attached to view.
class VIEWTOOLMGR {
	public:

		//! Construct for specified view.
		explicit VIEWTOOLMGR (
			GRE::UI::VIEW_CANVAS *view
			);

		~VIEWTOOLMGR ();

		//! Activate tool by "hotkey" code.
		//! @return true if found matching tool, false if not.
		bool ActivateByHotKey (
			MGUI::KEYCODE keycode
			);

		//! Add tool to manager.
		void Add (
			VIEWTOOL *tool
			);

		//! Add toggles to menu.
		void AddMenuItems (
			MGUI::MENU& Menu
			);

		//! Add toggles to toolbar.
		void AddToolBarItems (
			MGUI::CTRL_TOOLBAR& ToolBar
			);

		//! Delete all "script" tools.
		void DeleteAllScripts (
			bool RemoveUI						//!< Remove user interface controls if any
			);

		//! Find tool using builtin ID.
		VIEWTOOL* Find (
			VIEWTOOLID ViewToolID
			) const;

		//! Find tool using name.
		VIEWTOOL* FindByName (
			const MISTRING& name
			) const;

		//! Get active tool.
		VIEWTOOL* GetActive (
			) const { return (m_ActiveTool); }

		//! Get default tool.
		VIEWTOOL* GetDefault (
			) const;

		//! Get canvas view.
		VIEW_CANVAS* GetView (
			) const { return (m_view); }

		//! Determine if specified tool is active.
		bool IsActive (
			const VIEWTOOL* tool
			) const { return (tool != 0 && tool == m_ActiveTool); }

		//! Resume active tool.
		void Resume (
			);

		//! Set tool as active in manager.
		//! For internal use only, does not activate tool itself.
		void SetActive (
			VIEWTOOL *tool
			) { m_ActiveTool = tool; }

		//! Suspend active tool.
		void Suspend (
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		friend class VIEWTOOL;
		typedef MILIST<VIEWTOOL*> TOOLLIST;
		VIEW_CANVAS *m_view;
		TOOLLIST m_ToolList;								// List of tools
		VIEWTOOL *m_ActiveTool;							// Tool currently in use

		// Unimplemented.
		VIEWTOOLMGR (const VIEWTOOLMGR&);
		VIEWTOOLMGR& operator= (const VIEWTOOLMGR&);
		#endif
	};

//===================================================================================================================

//! Base class for tool attached to view.
class VIEWTOOL {
	public:

		//! Destructor
		virtual ~VIEWTOOL ();

		//! Activate this tool.
		void Activate (
			);

		//! Add toggle to menu.
		void AddMenuItem (
			MGUI::MENU& Menu
			);

		//! Add toggle to toolbar.
		void AddToolBarItem (
			MGUI::CTRL_TOOLBAR& ToolBar
			);

		//! Deactivate this tool.
		void Deactivate (
			);

		//! Get "hotkey" KEYCODE.
		MGUI::KEYCODE GetHotKeyCode (
			) const { return (m_HotKeyCode); }

		//! Get icon ID.
		ICONID GetIconID (
			) const { return (m_iconid); }

		//! Get name for menu and ToolTip.
		const MISTRING& GetName (
			) const { return (m_name); }

		//! Get parent for tool dialogs.
		MDLGPARENT GetDlgParent (
			) const { return (GetView()->GetDlgParent()); }

		//! Get view.
		GRE::UI::VIEW_CANVAS* GetView (
			) const { return (m_pMgr->GetView()); }

		//! Get viewable object.
		VIEWABLE* GetViewable (
			) const { return (GetView()->GetViewable()); }

		//! Get ID for builtin tool.
		const VIEWTOOLID GetViewToolID (
			) const { return (m_id); }

		//! Determine if tool is currently active.
		bool IsActive (
			) const;

		//! Determine if tool is currently disabled.
		bool IsDisabled (
			) const { return (m_IsDisabled); }

		//! Determine if tool is user-added script.
		bool IsScript (
			) const { return (m_IsScript); }

		//! Determine if tool is "suspended".
		bool IsSuspended (
			) const { return (m_SuspendCount > 0); }

		//! Remove user interface controls.
		void RemoveUI (
			);

		//! Resume tool (after screen update)
		void Resume (
			);

		//! Set whether items on toolbar and menu are enabled.
		void SetEnabled (
			bool enable
			);

		//! Suspend tool (during screen update).
		void Suspend (
			);

	protected:

		enum TOOLSTATE {
			STATE_NotInitialized = 0,
			STATE_Inactive,
			STATE_Active,
			STATE_NoLayers,
			STATE_Suspended
			};

		VIEWTOOL (
			GRE::UI::VIEW_CANVAS& view,
			const MISTRING& name,
			ICONID iconid,
			ICID icidMenu,
			ICID icidToolBar,
			VIEWTOOLID toolid = VIEWTOOLID_Unspecified,
			char HotKey = 0
			);

		VIEWTOOL (
			GRE::UI::VIEW_CANVAS& view,
			const MISTRING& name,
			ICONID iconid,
			VIEWTOOLID toolid = VIEWTOOLID_Unspecified,
			char HotKey = 0
			);

		//! Get current tool state.
		TOOLSTATE GetState (
			) const { return (m_state); }

		void SetIsScript (
			) { m_IsScript = true; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef VIEWTOOL THISCLASS;

		GRE::UI::VIEWTOOLMGR *m_pMgr;
		GRE::UI::VIEWTOOLID m_id;
		MISTRING m_name;
		ICONID m_iconid;
		TOOLSTATE m_state;
		bool m_IsDisabled;
		bool m_IsScript;
		ICID m_icidMenu;								// ID for tool menu item
		ICID m_icidToolBar;							// ID for toolbar item
		MGUI::KEYCODE m_HotKeyCode;
		MGUI::MENU m_Menu;
		MGUI::MENU::ITEMHND m_ItemHndMenu;
		MGUI::CTRL_TOOLBAR *m_pToolBar;
		UINT32 m_ItemHndToolBar;
		UINT32 m_SuspendCount;

		bool CanActivate () const;
		void OnToggle (MGUI::ID, bool ToggleOn);

		// Unimplemented.
		VIEWTOOL (const VIEWTOOL&);
		VIEWTOOL& operator= (const VIEWTOOL&);
		#endif

		//! \name Overridables
		//!@{

		virtual bool v_Activate () = 0;
		virtual void v_Deactivate () = 0;
		virtual bool v_Initialize ();
		virtual void v_Resume ();
		virtual void v_Suspend ();

		//!@}
	};

//===================================================================================================================

}	// End namespace UI
}	// End namespace GRE

#endif	// INC_GRE_UIVIEWTOOL_H
