/**
 * \file mgui/dlggetobjectitem.h
 * \brief MGUI::DLG_GETOBJECT::ITEM definition
 *
 * \if NODOC
 * $Id: dlggetobjectitem.h_v 1.31 2013/05/28 16:38:09 bcolombe Exp $
 *
 * $Log: dlggetobjectitem.h_v $
 * Revision 1.31  2013/05/28 16:38:09  bcolombe
 * Add SetName from objitem
 *
 * Revision 1.30  2012/12/17 21:27:39  bcolombe
 * Add success,failure count
 *
 * Revision 1.29  2012/09/11 16:11:56  bcolombe
 * Support fusion table
 *
 * Revision 1.28  2012/04/03 22:07:39  bcolombe
 * Fix for download err
 *
 * Revision 1.27  2012/01/27 22:14:03  bcolombe
 * Added support for terrain tileset
 *
 * Revision 1.26  2011/11/09 14:48:42  mju
 * Remove redundant hdrs.
 *
 * Revision 1.25  2011/03/14 18:38:56  bcolombe
 * Add TYPE_Geomashup
 *
 * Revision 1.24  2011/02/11 22:36:48  bcolombe
 * Add order
 *
 * Revision 1.23  2011/02/07 15:19:39  bcolombe
 * Add SetSelectable
 *
 * Revision 1.22  2010/10/04 21:21:26  bcolombe
 * Add getinistring, setfroministring
 *
 * Revision 1.21  2010/09/17 20:58:41  bcolombe
 * Change for getextents/crs
 *
 * Revision 1.20  2010/09/16 22:52:42  bcolombe
 * Change to getexents, getcrs methods
 *
 * Revision 1.19  2010/09/02 22:59:13  bcolombe
 * Added delegate for loading web layers
 *
 * Revision 1.18  2010/08/27 17:32:41  bcolombe
 * Add constructor with name and objitem
 *
 * Revision 1.17  2010/08/24 17:39:05  bcolombe
 * Add optional param to GetDisplayPath
 *
 * Revision 1.16  2010/08/19 21:31:46  bcolombe
 * Added GetDisplayLocation
 *
 * Revision 1.15  2010/08/18 22:38:38  bcolombe
 * Make subtype INT32 for use with proprietary ids
 *
 * Revision 1.14  2010/08/16 18:51:28  bcolombe
 * Make operator== const
 *
 * Revision 1.13  2010/08/11 22:32:11  bcolombe
 * Added item for weblayers
 *
 * Revision 1.12  2010/02/17 22:41:46  mju
 * Add dimension, crs, extent, etc values.
 *
 * Revision 1.11  2010/02/17 18:34:18  mju
 * Add getDescription.
 *
 * Revision 1.10  2009/11/23 21:06:23  mju
 * Make setName public.
 *
 * Revision 1.9  2009/11/20 15:59:49  mju
 * Make m_pnetRes mutable.
 *
 * Revision 1.8  2009/11/13 15:36:55  mju
 * Add equality/inequality operators.
 *
 * Revision 1.7  2009/11/12 17:21:58  mju
 * Add getIconID.
 *
 * Revision 1.6  2009/11/09 15:18:48  mju
 * Add isSelectable parm for item ctor from filepath.
 *
 * Revision 1.4  2009/10/30 18:07:37  mju
 * Don't use default type parms.
 *
 * Revision 1.3  2009/10/28 21:28:38  mju
 * Make getObjectState non-inline.
 *
 * Revision 1.2  2009/10/23 19:46:30  mju
 * Add getObjItemToChange and resetExtValues.
 *
 * Revision 1.1  2009/10/22 21:50:55  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MGUI_DLGGETOBJECTITEM_H
#define	INC_MGUI_DLGGETOBJECTITEM_H

#ifndef	INC_MGUI_DLGGETOBJECTDLG_H
#include <mgui/dlggetobjectdlg.h>
#endif

#ifndef	INC_MI32_SPATREF_H
#include <mi32/spatref.h>
#endif

#ifndef	INC_MI32_DATETIME_H
#include <mi32/datetime.h>
#endif

#ifndef	INC_MI32_ELEMTYPE_H
#include <mi32/elemtype.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
class XMLNODE;
namespace RVC {
	class GEOMETRIC;
	}
#endif

namespace MGUI {
//=====================================================================================================================
//		DLG_GETOBJECT::ITEM definition - Begin

class DLG_GETOBJECT::ITEM {
	public:
		typedef fastdelegate::FastDelegate<ERRVALUE(ITEM&)> DELEGATE_ONLOADWEBLAYERS;

		enum TYPE {
			TYPE_Invalid,
			TYPE_Computer,
			TYPE_NetRoot,			// Root of network
			TYPE_NetServer,		// Network server
			TYPE_NetShare,			// Network shared directory
			TYPE_Volume,			// Drive / Volume
			TYPE_Folder,			// Local directory
			TYPE_File,				// File
			TYPE_RVC_File,			// RVC file
			TYPE_RVC_Folder,		// RVC folder
			TYPE_RVC_Object,		// RVC object
			TYPE_WebLayerGroup,		// Web Layer Group
			TYPE_RemoteTileset,		// Remote TSD
			TYPE_RemoteKML,			// Remote KML
			TYPE_MyMaps,			// MyMaps
			TYPE_WMS,				// WMS
			TYPE_ArcIMS,			// ArcIMS
			TYPE_Geomashup,			// Geomashup file
			TYPE_TerrainTileset,	// Terrain tileset
			TYPE_FusionTable,		// Fusion Table
			TYPE_COUNT
			};

		enum OBJECTSTATE {
			OBJECTSTATE_Unknown = 0,
			OBJECTSTATE_Valid = 1,
			OBJECTSTATE_IsInvalidChild,					//< Is not a valid child for its parent
			OBJECTSTATE_HasInvalidChild,					//< Has invalid children, object itself may be valid
			OBJECTSTATE_IsInvalidSingleton,				//< Is not a singleton for parent as required
			OBJECTSTATE_HasInvalidSingleton,				//< Has multiple children of same type where singleton required
			OBJECTSTATE_ObjectHasUnlocatableLink,		//< Object linked to unlocatable resource
			OBJECTSTATE_ChildHasUnlocatableLink,		//< Has descendent linked to unlocatable resource
			OBJECTSTATE_IsOutdatedChild,					//< Is outdated child relative to its parent
			OBJECTSTATE_HasOutdatedChild,					//< Has outdated child
			};

		// Default constructor.
		ITEM ();

		// Copy constructor.
		ITEM (const ITEM& rhs);

		// Construct with specified name and TYPE.
		ITEM (const MISTRING& name, TYPE type);

		// Construct with specified name and RVC::OBJTYPE.
		ITEM (const MISTRING& name, RVC::OBJTYPE rvctype);

		// Construct from FILEPATH and TYPE.
		ITEM (const FILEPATH& filepath, TYPE type, bool IsSelectable);

		// Construct named FILEPATH.
		ITEM (const MISTRING& name, const FILEPATH& filepath, TYPE type);

		// Construct from RVC::OBJITEM.
		ITEM (const RVC::OBJITEM& objitem, bool IsBrowsable = false, bool IsSelectable = false);

		// Construct from RVC::OBJITEM.
		ITEM (const MISTRING& name, const RVC::OBJITEM& objitem, bool IsBrowsable = false, bool IsSelectable = false);

		// Construct from MINETRES.
		ITEM (const MINETRES& netres);

		// Construct named location.
		ITEM (const MISTRING& name, const MISTRING& location, TYPE type);

		// Construct from XMLNODE
		ITEM (XMLNODE* node);

		// Destructor.
		~ITEM ();

		// Assignment operator.
		ITEM& operator= (
			const ITEM& rhs
			);

		// Comparison for equality.
		bool operator== (
			const DLG_GETOBJECT::ITEM& rhs
			) const { return (IsEqual(rhs)); }

		// Comparison for inequality.
		bool operator!= (
			const DLG_GETOBJECT::ITEM& rhs
			) { return (!IsEqual(rhs)); }

		// Build network item list.
		ERRVALUE BuildNetItemList (
			ITEMLIST& ItemList
			) const;

		//! Get CRS
		const SPATREF::COORDREFSYS& GetCoordRefSys (
			) const {return m_CRS;}

		//! Get item description.
		const MISTRING& GetDescription (
			) const { return (m_description); }

		//! Get failure count for wms/arcims item
		UINT32 GetFailureCount (
			) const { return (m_FailureCount); }

		//! Get success count for wms/arcims item
		UINT32 GetSuccessCount (
			) const { return (m_SuccessCount); }

		//! Get dimension value for sorting.
		const MISTRING& GetDimensionString (
			) const { return (m_DimensionString); }

		//! Get dimension value for sorting.
		INT64 GetDimensionValue (
			) const { return (m_DimensionValue); }

		//! Get extents in CRS.
		const DRECT3D& GetExtentInCRS (
			) const {return m_ExtentInCRS;}

		//! Get icon ID
		ICONID GetIconID (
			) const;

		const FILEPATH& GetFilePath (
			) const { return (m_pObjItem->GetFilePath()); }

		// Get item name.
		const MISTRING& GetName (
			) const;

		// Get item location for display
		void GetDisplayLocation (
			MISTRING& location,
			bool bStripToFolder = true
			) const;

		// Get item location.
		const MISTRING& GetLocation (
			) const {return (m_location); }

		MINETRES* GetNetRes (
			) const { return (m_pNetRes); }

		XMLNODE* GetNode (
			bool bLoadIfNeeded = true
			);

		// Get object state.
		OBJECTSTATE GetObjectState (
			) const;

		// Get RVC object item information.
		// @return Pointer to RVC::OBJITEM or NULL if does not refer to RVC object.
		const RVC::OBJITEM* GetObjItem (
			) const { return (m_pObjItem); }

		// Get RVC object item information to change.
		RVC::OBJITEM& GetObjItemToChange (
			) { return (*m_pObjItem); }

		INT32 GetOrder (
			) const { return (m_order); }

		const DPOINT3D& GetResolutionInCRS (
			) const { return (m_ResolutionInCRS); }

		RVC::OBJTYPE GetRVCType (
			) const { return (m_rvctype); }

		// Get item short name.
		const RVC::OBJECTNAME& GetShortName (
			) const;

		INT64 GetSizeInBytes (
			) const { return (m_SizeInBytes); }

		INT32 GetSubType (
			) const { return (m_subtype); }

		// Get date/time created.
		const DATETIME& GetTimeCreated (
			) const { return (m_TimeCreated); }

		// Get date/time modified.
		const DATETIME& GetTimeModified (
			) const { return (m_TimeModified); }

		// Get item type.
		TYPE GetType (
			) const { return (m_type); }

		bool IsBrowsable (
			) const { return (m_IsBrowsable); }

		bool IsEqual (
			const ITEM& rhs
			) const;

		bool IsSelectable (
			) const { return (m_IsSelectable); }

		void ResetExtValues (
			) { m_ExtValuesValid = EXTVALUE_None; }

		void SetDelegateOnLoadWebLayers (
			DELEGATE_ONLOADWEBLAYERS delegate
			) { m_DelegateOnLoadWebLayers = delegate; }

		void SetCRS (const SPATREF::COORDREFSYS& crs) {m_CRS = crs;}

		void SetExtentInCRS (const DRECT2D& extents) { m_ExtentInCRS = extents;}

		// Set name.
		void SetName (const MISTRING& name);

		void SetName (const RVC::OBJITEM& objitem);

		void SetSuccessCount (UINT32 SuccessCount) {m_SuccessCount = SuccessCount;}

		void SetFailureCount (UINT32 FailureCount) {m_FailureCount = FailureCount;}

		void SetNode (XMLNODE* node) { m_node = node; }

		void SetSelectable (bool IsSelectable) { m_IsSelectable = IsSelectable; }

		void SetupExtValues (EXTVALUES ExtValues);

		//! Get ini string, returns false if failed
		bool GetINIString (MISTRING& inistring) const;

		//! Set from ini string, returns false if failed
		bool SetFromINIString (const MISTRING& inistring);

	protected:

		void SetFilePath (
			const FILEPATH& filepath
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MISTRING m_name;
		MISTRING m_location;
		MISTRING m_description;
		TYPE m_type;
		RVC::OBJTYPE m_rvctype;
		INT32 m_subtype;
		bool m_IsBrowsable;
		bool m_IsSelectable;
		UINT32 m_ExtValuesValid;
		UINT32 m_FailureCount;
		UINT32 m_SuccessCount;
		RVC::OBJITEM *m_pObjItem;
		mutable MINETRES *m_pNetRes;
		mutable OBJECTSTATE m_ObjectState;
		XMLNODE* m_node;
		INT32 m_order;
		DATETIME m_TimeCreated;
		DATETIME m_TimeModified;
		INT64 m_SizeInBytes;
		INT64 m_DimensionValue;
		MISTRING m_DimensionString;
		SPATREF::COORDREFSYS m_CRS;
		DRECT3D m_ExtentInCRS;
		DPOINT3D m_ResolutionInCRS;
		DELEGATE_ONLOADWEBLAYERS m_DelegateOnLoadWebLayers;

		ELEMTYPE AddDimensionElemCount (RVC::GEOMETRIC& ObjGeometric, ELEMTYPE ElemType, ELEMTYPE PrevType);
		// Set description.
		void SetDescription (const MISTRING& description);
		#endif
	};

//		DLG_GETOBJECT::ITEM definition - End
//=====================================================================================================================
}	// End namespace MGUI

#endif	// INC_MGUI_DLGGETOBJECTITEM_H
