/**
 * \file mgui/toolcanvas.h
 * \brief MGUI::TOOL_CANVAS definitions.
 *
 * \if NODOC
 * $Id: toolcanvas.h_v 1.20 2012/01/12 18:12:15 mju Exp $
 *
 * $Log: toolcanvas.h_v $
 * Revision 1.20  2012/01/12 18:12:15  mju
 * Add stereoCursorDraw/Restore methods.
 *
 * Revision 1.19  2011/12/30 17:17:26  mju
 * Remove getToolDrawingContext.
 *
 * Revision 1.18  2011/12/08 22:23:15  mju
 * Add getToolDrawingContext.
 *
 * Revision 1.17  2011/12/01 22:30:08  mju
 * Add getZView and virtual overridable.
 *
 * Revision 1.13  2011/03/09 22:39:39  mju
 * Move all private stuff to private class.
 *
 * Revision 1.12  2011/02/25 18:59:14  mju
 * More gadget support methods.
 *
 * Revision 1.10  2010/10/25 14:36:31  mju
 * Add "hasmore" to v_onPointerMove.
 *
 * Revision 1.9  2008/01/25 15:31:16  mju
 * Add v_onPointerEnter/Leave overrides.
 *
 * Revision 1.6  2007/11/30 17:28:01  mju
 * Make overrides protected so can call from derived.
 *
 * Revision 1.1  2003/01/08 22:56:58  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MGUI_TOOLCANVAS_H
#define	INC_MGUI_TOOLCANVAS_H

#ifndef	INC_MGUI_CANVAS_H
	#include <mgui/canvas.h>
#endif

#ifndef	INC_MI32_STEREOVIEW_H
	#include <mi32/stereoview.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class ZVIEW;
namespace MGUI {
	class GADGET;
	class TOOL_BASE;
	}
#endif


namespace MGUI {
//=====================================================================================================================

// Container canvas for graphic tools
class TOOLCANVAS : public CANVAS {

	public:

		//! Default constructor.
		TOOLCANVAS ();
		
		//! Destructor.					
		virtual ~TOOLCANVAS ();

		//! Draw all gadgets on specified device type.
		void DrawGadgets (
			DEVICETYPE DeviceType
			);

		//! Get pointer to active gadget.
		GADGET* GetActiveGadget () const;

		//! Get pointer to currently active graphic tool.
		TOOL_BASE* GetActiveTool () const;

		//! Get ZVIEW for surface value access and stereo drawing.
		//! @return Pointer to ZVIEW if available, NULL if not.
		const ZVIEW* GetZView () const
			{ return (v_GetZView()); }

		//! Resume active graphic tool.
		void GraphicToolResume () const;

		//! Suspend active graphic tool.
		void GraphicToolSuspend () const;

		//! Place gadgets on canvas.
		void PlaceGadgets ();

		//! Set gadget margin.
		void SetGadgetMargin (
			UINT16 margin					//!< Margin between gadgets and canvas edge in pixels
			);

		//! Set gadget spacing.
		void SetGadgetSpacing (
			UINT16 spacing					//!< Minimum spacing between gadgets in pixels
			);

		//! Draw stereo cursor using absolute Z value.
		//! Does not restore under previous stereo cursor if any.
		void StereoCursorDrawAbsolute (
			const DPOINT3D& ptCanvas3D			//!< Point with X/Y in canvas coordinates and Z value set
			);

		//! Draw stereo cursor on surface.
		//! Does not restore under previous stereo cursor if any.
		void StereoCursorDrawOnSurface (
			MGUI::POINT ptCanvas					//!< Point with X/Y in canvas coordinates
			);

		//! Restore under stereo cursor if was previously drawn.
		void StereoCursorRestore ();

	protected:

		//! Called when cursor "capture" is changed (lost).
		//! Override must call base for proper behavior.
		virtual void v_OnCaptureChanged ();

		//! Called when key "pressed".
		//! Override must call base for proper behavior if does not handle key itself.
		//! @return 'true' if key handled, 'false' if not.
		virtual bool v_OnKey (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			MGUI::KEYCODE keycode,				//!< Processed key code, equivalent to MIUNICODE for alphanumeric keys.
			MGUI::KEYSTATE keystate				//!< Modifier key state
			);

		//! Called when canvas needs to be redrawn, do not call directly except from derived implementation.
		//! Override must call base for proper behavior.
		//! This is called after the buffer is updated to the screen.  If no overlay and not unbuffered
		//! then this method does not need to be overridden.
		virtual void v_OnPaint (
			MGD::CONTEXT *gc						//!< Drawing context for unbuffered/overlay.
			);

		//! Called when pointer button pressed.
		//! Override must call base for proper behavior.
		virtual void v_OnPointerButtonDown (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Button that was pressed
			KEYSTATE keystate						//!< Modifier key state at time of event
			);

		//! Called when pointer button released.
		//! Override must call base for proper behavior.
		virtual void v_OnPointerButtonUp (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Button that was released
			KEYSTATE keystate						//!< Modifier key state at time button was originally pressed
			);

		//! Called when pointer enters canvas area.
		//! Override must call base for proper behavior.
		virtual void v_OnPointerEnter ();

		//! Called when pointer leaves canvas area.
		//! Override must call base for proper behavior.
		virtual void v_OnPointerLeave ();

		//! Called when pointer is moved.
		//! Override must call base for proper behavior.
		virtual void v_OnPointerMove (
			MGUI::POINT point,					//!< Point in canvas coordinates where button event occurred
			POINTERBUTTON button,				//!< Current button state
			KEYSTATE keystate,					//!< Modifier key state at time of event
			bool hasmore							//!< Indicates whether more "move" events follow with same button/key state
			);

		//! Called when canvas size changes.
		//! Override must call base for proper behavior.
		virtual void v_OnSize (
			int width,								//!< New canvas width in pixels
			int height								//!< New canvas height in pixels
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef CANVAS BASECLASS;

		class PRIV;
		PRIV *m_pPriv;

		friend class TOOL_BASE;
		friend class GADGET;
		#endif // GENERATING_DOXYGEN_OUTPUT

		void GadgetAttach (GADGET * const pGadget);
		void GadgetDetach (GADGET * const pGadget);
		void SetActiveTool (TOOL_BASE* pTool);
		void ToolAttach (TOOL_BASE *const pTool);
		void ToolDetach (TOOL_BASE *const pTool);

		//! Get ZVIEW for surface value access and stereo drawing.
		//! Default implementation returns NULL.
		virtual const ZVIEW* v_GetZView () const;
	};

//=====================================================================================================================

}	// End namspace MGUI

#endif	// INC_MGUI_TOOLCANVAS_H
