/**
 *	\file mi32/drawgen.h
 *	\brief Declarations for "generic" drawing functions
 *
 *	\if NODOC
 *	$Log: drawgen.h_v $
 *	Revision 1.69  2012/12/10 18:55:35  mju
 *	Add drawLine/SegmentThick.
 *
 *	Revision 1.68  2012/12/10 14:00:26  mju
 *	Remove remaining obsolete defns.
 *
 *	Revision 1.67  2012/12/06 15:00:38  mju
 *	Remove unused fn prototypes.
 *
 *	Revision 1.66  2012/12/05 14:35:04  mju
 *	Add fillSub/Region.
 *
 *	Revision 1.65  2012/12/04 14:54:21  mju
 *	Add flag_NoLineSpans.
 *	Remove drawLineGen.
 *
 *	Revision 1.64  2012/12/04 13:38:51  mju
 *	Move drawgen class to dll.
 *
 *	Revision 1.63  2012/12/03 17:02:45  mju
 *	Add drawgen class with delegates.
 *	Remove drawGenCircleFilled as now in class.
 *
 *	Revision 1.62  2012/10/01 15:47:11  mju
 *	Add drawGenSeededFill and drawgenflags.
 *
 *	\endif
**/

#ifndef INC_MI32_DRAWGEN_H
#define INC_MI32_DRAWGEN_H

#ifndef	INC_MI32_DELEGATE_H
#include <mi32/delegate.h>
#endif


#ifdef GEOMDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class POLYLINE;
class POLYLINELIST;
class REGION2D;
class SUBREGION2D;
#endif

#if defined(__cplusplus)
extern "C" {
#endif

//=====================================================================================================================

//! Generic drawing/tracing operations.
//! No inherent clipping or range testing is performed, however all coordinates are expected to be in the range of
//! a 32-bit signed integer (INT32).  In addition, because operations are performed on an integer grid, use of
//! very large ranges may adversely affect performance and/or memory usage.
//! Typically at least a "draw span" delegate must be specified on construction.
class CLASSLIBEXPORT DRAWGEN {
	public:

		//! Flags for various operations.
		enum FLAGS {
			FLAG_None =				0,
			FLAG_Closed =			0x0001,	//!< Close polylines
			FLAG_4Connected =		0x0002,	//!< Use 4-connected lines
			FLAG_8Connected =		0x0004,	//!< Use 8-connected lines
			FLAG_NoLineSpans =	0x0008,	//!< Don't draw multi-point spans for line/segment
			};

		//! Delegate for drawing span, parameters are X1, X2, Y.
		typedef fastdelegate::FastDelegate<ERRVALUE(INT32,INT32,INT32)> DELEGATE_DRAWSPAN;

		//! Delegate for testing single point, parameters are X, Y, return TRUE if included, FALSE if not, < 0 error.
		typedef fastdelegate::FastDelegate<int(INT32,INT32)> DELEGATE_TESTPOINT;

		//! Construct with delegate to draw span.
		DRAWGEN (
			DELEGATE_DRAWSPAN DrawSpan					//!< Delegate to draw span
			);

		//! Construct with delegates to draw span and test point.
		DRAWGEN (
			DELEGATE_DRAWSPAN DrawSpan,				//!< Delegate to draw span
			DELEGATE_TESTPOINT TestPoint				//!< Delegate to test single point
			);

		//! Construct with only delegate to test point.
		//! If this constructor is used, only Trace... methods may be called
		DRAWGEN (
			DELEGATE_TESTPOINT TestPoint				//!< Delegate to test single point
			);

		//! Draw line from POLYLINE.
		//! Note that unless FLAG_NoLineSpans is specified, multiple points with the same y coordinate may be combined in a single call
		//! to the "draw span" delegate.  If the sequence of points in important, FLAG_NoLineSpans must be specified.
		ERRVALUE DrawLine (
			const POLYLINE& polyline,					//!< Line, may be closed
			FLAGS flags = FLAG_None						//!< Flags (FLAG_NoLineSpans, FLAG_4Connected, FLAG_8Connected, default is 8-connected)
			);

		//! Draw thick line.
		ERRVALUE DrawLineThick (
			const POLYLINE& polyline,					//!< Line, may be closed
			double thickness,								//!< Thickness in pixels, if < 1.5 will be same as DrawGenLine
			FLAGS flags = FLAG_None						//!< Flags (FLAG_4Connected, default is 8-connected)
			);

		//! Draw line segment.
		//! Note that unless FLAG_NoLineSpans is specified, multiple points with the same y coordinate may be combined in a single call
		//! to the "draw span" delegate.  If the sequence of points in important, FLAG_NoLineSpans must be specified.
		ERRVALUE DrawSegment (
			const DPOINT2D& pt1,							//!< Starting point
			const DPOINT2D& pt2,							//!< Ending point
			FLAGS flags = FLAG_None						//!< Flags (FLAG_NoLineSpans, FLAG_4Connected, FLAG_8Connected, default is 8-connected)
			);

		//! Draw thick line segment.
		ERRVALUE DrawSegmentThick (
			const DPOINT2D& pt1,							//!< Starting point
			const DPOINT2D& pt2,							//!< Ending point
			double thickness,								//!< Thickness in pixels, if < 1.5 will be same as DrawGenLine
			FLAGS flags = FLAG_None						//!< Flags (FLAG_NoLineSpans, FLAG_4Connected, FLAG_8Connected, default is 8-connected)
			);

		//! Draw filled circle.
		ERRVALUE FillCircle (
			INT32 x, 										//!< X coordinate of center
			INT32 y, 										//!< Y coordinate of center
			INT32 radius 									//!< Radius in pixels
			);

		//! Draw filled polygon.
		//! If there is any overlap, the even/odd rule is used.
		//! Requires approximately 24 bytes per input polygon vertex.
		//! Cells are filled if the center (x.5, y.5) is inside or on the border of the polygon.
		ERRVALUE FillPolygon (
			const POLYLINE& polyline,					//!< Polygon (assumed to be closed)
			FLAGS flags = FLAG_None						//!< Flags
			);

		//! Draw filled polygons.
		//! If there is any overlap, the even/odd rule is used.
		//! Requires approximately 24 bytes per input polygon vertex.
		//! Cells are filled if the center (x.5, y.5) is inside or on the border of the polygon.
		ERRVALUE FillPolygons (
			const POLYLINELIST& polylist,				//!< Polygons (assumed to be closed)
			FLAGS flags = FLAG_None						//!< Flags
			);

		//! Draw filled region.
		//! Cells are filled if the center (x.5, y.5) is inside or on the boundary of the region.
		ERRVALUE FillRegion (
			const REGION2D& region,						//!< Region to fill
			FLAGS flags = FLAG_None						//!< Flags
			);

		//! Seeded-fill drawing.
		//! The "Test" delegate may be called on already-filled points and must return FALSE in that case.
		//! Supports both 4-connect and 8-connected fills, will be 4-connected unless FLAG_8Connected is specified.
		//! Must specify valid DrawSpan and TestPoint delegates on construction.
		ERRVALUE FillSeeded (
			INT32 SeedX,									//!< Seed point X coordinate 
			INT32 SeedY, 									//!< Seed point Y coordinate
			FLAGS flags = FLAG_None						//!< Flags (FLAG_8Connected, default is 4-connected fill)
			);

		//! Draw filled subregion.
		//! Requires approximately 24 bytes per input polygon vertex.
		//! Cells are filled if the center (x.5, y.5) is inside or on the border of the subregion.
		ERRVALUE FillSubRegion (
			const SUBREGION2D& subregion,				//!< Subregion to fill
			FLAGS flags = FLAG_None						//!< Flags
			);

		//! General boundary trace.
		//! Must specify valid TestPoint delegate on construction.
		ERRVALUE TraceBoundary (
			const LPOINT2D& ptStart,					//!< Starting point inside area to be traced around
			POLYLINE& polyline,							//!< Polygon returned
			FLAGS flags = FLAG_None						//!< Flags (FLAG_8Connected, default is 4-connected)
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		struct SEED_FILL {
			INT32 left;
			INT32 right;
			INT32 y;

			SEED_FILL (INT32 leftinput, INT32 rightinput, INT32 yinput) {left = leftinput; right = rightinput; y = yinput;}
			};

		DELEGATE_DRAWSPAN m_DelegateDrawSpan;
		DELEGATE_TESTPOINT m_DelegateTestPoint;
		#endif
	};

//=====================================================================================================================

#if defined(__cplusplus)
}
#endif

#undef	CLASSLIBEXPORT

#endif
