/**
 * \file mi32/hdf4.h
 * \brief HDF4 class definitions
 *
 * \if NODOC
 * $Id: hdf4.h_v 1.6 2011/10/26 12:40:13 mju Exp $
 *
 * $Log: hdf4.h_v $
 * Revision 1.6  2011/10/26 12:40:13  mju
 * Remove redundant hdr.
 *
 * Revision 1.5  2007/05/07 13:30:23  mju
 * Use fixedstrchar hdr.
 *
 * Revision 1.2  2002/05/20 16:17:43  mju
 * Add FIELDINFO and RECORD defns.
 *
 * Revision 1.1  2002/05/17 21:55:55  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MI32_HDF4_H
#define	INC_MI32_HDF4_H

#ifndef INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifndef INC_MI32_FIXEDSTRCHAR_H
#include <mi32/fixedstrchar.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
class FILEPATH;
class BITSET;
#endif


//! HDF4 object types.
enum HDF4_OBJTYPE {
	HDF4_OBJTYPE_Group =		0,		//!< HDF VG (Vgroup)
	HDF4_OBJTYPE_Array =		1,		//!< HDF SD (Scientific Data)
	HDF4_OBJTYPE_Table =		2,		//!< HDF VD (Vdata)
	HDF4_OBJTYPE_Raster =	3,		//!< HDF GR (General Raster)
	HDF4_OBJTYPE_NUMTYPES
	};

//! HDF4 value types.
enum HDF4_VALTYPE {
	HDF4_VALTYPE_None =		0,
	HDF4_VALTYPE_UCHAR8 =	3,		//!< String
	HDF4_VALTYPE_CHAR8 =		4,		//!< String
	HDF4_VALTYPE_FLOAT32 =	5,
	HDF4_VALTYPE_FLOAT64 =	6,
	HDF4_VALTYPE_INT8 =		20,
	HDF4_VALTYPE_UINT8 =		21,
	HDF4_VALTYPE_INT16 =		22,
	HDF4_VALTYPE_UINT16 =	23,
	HDF4_VALTYPE_INT32 =		24,
	HDF4_VALTYPE_UINT32 =	25,
	HDF4_VALTYPE_INT64 =		26,
	HDF4_VALTYPE_UINT64 =	27,
	};


#ifndef GENERATING_DOXYGEN_OUTPUT
class HDF4_FILE_PRIV;
class HDF4_RECORD;
#endif //!< GENERATING_DOXYGEN_OUTPUT

class HDF4_NAME : public FIXEDSTRCHAR<256> {
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Get non-const pointer to string.
		char * GetStr () { return (m_String); }

		friend class HDF4_GROUP;
		friend class HDF4_ARRAY;
		friend class HDF4_RASTER;
		friend class HDF4_TABLE;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

class HDF4_FILE {
	public:

		//! Default constructor.
		HDF4_FILE (
			): m_pPriv(0) { }

		//! Copy constructor.
		HDF4_FILE (
			const HDF4_FILE& rhs
			);

		//! Destructor.
		~HDF4_FILE (
			);

		//! Assignment
		HDF4_FILE& operator= (
			const HDF4_FILE& rhs
			);

		//! Close currently open file.
		void Close (
			);

		//! Determine if file is currently open.
		bool IsOpen (
			) const { return (m_pPriv != 0); };

		//! Open file.
		ERRVALUE Open (
			const FILEPATH& filepath
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		HDF4_FILE_PRIV *m_pPriv;

		friend class HDF4_GROUP;
		friend class HDF4_ARRAY;
		friend class HDF4_RASTER;
		friend class HDF4_TABLE;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_ATTRIB {
	public:

		//! Default constructor.
		HDF4_ATTRIB (
			);

		//! Get number of attribute values if numeric or number of characters if string.
		int GetCount (
			) const { return (m_count); }

		//! Get attribute name.
		const char *GetName (
			) const { return (m_name); }

		//! Get attribute value as a string, suitable for character attribute types.
		ERRVALUE GetString (
			SIMPLE_ARRAY<char>& string
			) const;

		//! Get attribute type.
		HDF4_VALTYPE GetType (
			) const { return (m_type); }

		//! Get attribute value as 'double', suitable for numeric attribute types.
		ERRVALUE GetValue (
			int pos,								//!< Attribute value position
			double& value						//!< Value returned
			) const;

		//! Determine if attribute is numeric.
		bool IsNumeric (
			) const { return (m_type != HDF4_VALTYPE_None && m_type != HDF4_VALTYPE_UCHAR8 && m_type != HDF4_VALTYPE_CHAR8); }

		//! Determine if attribute is a string.
		bool IsString (
			) const { return (m_type == HDF4_VALTYPE_UCHAR8 || m_type == HDF4_VALTYPE_CHAR8); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<char> m_buf;
		HDF4_NAME m_name;
		HDF4_VALTYPE m_type;
		int m_count;

		ERRVALUE Set (
			int type,
			int count
			);

		friend class HDF4_FILE;
		friend class HDF4_GROUP;
		friend class HDF4_ARRAY;
		friend class HDF4_RASTER;
		friend class HDF4_TABLE;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_OBJECT {
	public:

		//! Default constructor.
		HDF4_OBJECT (
			) { }

		//! Destructor.
		virtual ~HDF4_OBJECT (
			) = 0;

		//! Find attribute index by name.
		//! @return Attribute index, -1 if not found, < -1 if error.
		virtual int FindAttrib (
			const char *name
			) const = 0;

		//! Get specific attribute.
		virtual ERRVALUE GetAttrib (
			int index,
			HDF4_ATTRIB& attrib
			) const = 0;

		//! Get object name.
		virtual ERRVALUE GetName (
			HDF4_NAME& name						//!< Name returned
			) const = 0;

		//! Get number of attributes in this object.
		virtual int GetNumAttribs (
			) const = 0;

		//! Determine if object is currently open.
		bool IsOpen (
			) const { return (m_file.IsOpen()); };

	protected:

		HDF4_FILE m_file;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		HDF4_OBJECT (const HDF4_OBJECT&);
		HDF4_OBJECT& operator= (const HDF4_OBJECT&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_GROUP : public HDF4_OBJECT {
	public:

		//! Default constructor.
		HDF4_GROUP (
			);

		//! Destructor.
		virtual ~HDF4_GROUP (
			);

		//! Close group.
		void Close (
			);

		//! Find attribute index by name.
		//! @return Attribute index, -1 if not found, < -1 if error.
		virtual int FindAttrib (
			const char *name						//!< Attribute name to search for
			) const;

		//! Find object by type and path.
		//! Items in the path are separated by '/' with the last item being the object name.
		//! @return Index if found, -1 if not found, < -1 if error.
		int FindObject (
			HDF4_OBJTYPE type,					//!< HDF object type
			const char *path						//!< Path to search
			) const;

		//! Find first object having specified 'class' name.
		//! Items in the path are separated by '/' with the last item being the 'class' name.
		int FindObjectByClass (
			HDF4_OBJTYPE type,					//!< HDF object type (Group, Table only)
			const char *classname				//!< Object 'class' name
			) const;

		//! Get specific attribute.
		virtual ERRVALUE GetAttrib (
			int index,								//!< Attribute index within group
			HDF4_ATTRIB& attrib					//!< Attribute instance returned
			) const;

		//! Get 'class' name.
		ERRVALUE GetClass (
			HDF4_NAME& classname					//!< Class name returned
			) const;

		//! Get object name.
		virtual ERRVALUE GetName (
			HDF4_NAME& name						//!< Name returned
			) const;

		//! Get number of attributes in this object.
		virtual int GetNumAttribs (
			) const;

		//! Get array of object indices for group.
		//! Note that a specific object can be contained in multiple groups.
		ERRVALUE GetObjIdxs (
			HDF4_OBJTYPE objtype,				//!< Object type
			SIMPLE_ARRAY<INT32>& objidxs		//!< Object index array returned
			) const;

		//! Open group.
		ERRVALUE Open (
			const HDF4_FILE& file,				//!< Open HDF4 file
			int index								//!< Group index (reference number)
			);

		//! Open root group for file
		ERRVALUE OpenRoot (
			const HDF4_FILE& file
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_groupid;
		bool m_isroot;

		ERRVALUE CheckInSubGroup (HDF4_OBJTYPE objtype, BITSET& bitset) const;

		HDF4_GROUP (const HDF4_GROUP&);
		HDF4_GROUP& operator= (const HDF4_GROUP&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_ARRAY : public HDF4_OBJECT {
	public:

		//! Container for array information.
		class INFO;

		//! Default constructor.
		HDF4_ARRAY (
			);

		//! Destructor.
		virtual ~HDF4_ARRAY (
			);

		//! Close array.
		void Close (
			);

		//! Find attribute index by name.
		//! @return Attribute index, -1 if not found, < -1 if error.
		virtual int FindAttrib (
			const char *name						//!< Attribute name to search for
			) const;

		//! Get specific attribute.
		virtual ERRVALUE GetAttrib (
			int index,								//!< Attribute index within array
			HDF4_ATTRIB& attrib					//!< Attribute instance returned
			) const;

		//! Get array information.
		ERRVALUE GetInfo (
			INFO& info								//!< Information container returned.
			) const;

		//! Get object name.
		virtual ERRVALUE GetName (
			HDF4_NAME& name						//!< Name returned
			) const;

		//! Get number of attributes in this object.
		virtual int GetNumAttribs (
			) const;

		//! Open array.
		ERRVALUE Open (
			const HDF4_FILE& file,
			int index
			);

		//! Read array values into buffer.
		ERRVALUE Read (
			const INT32 *start,
			const INT32 *stride,
			const INT32 *count,
			void *buffer
			) const;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_arrayid;
		int m_numdims;
		int m_valuesize;

		HDF4_ARRAY (const HDF4_ARRAY&);
		HDF4_ARRAY& operator= (const HDF4_ARRAY&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//! HDF Array Information container.
class HDF4_ARRAY::INFO {
	public:

		//! Get dimension for specified axis.
		int GetDim (
			int axis
			) const { return (m_dims[axis]); }

		//! Get null value.
		double GetNullValue (
			) const { return (m_NullValue); }

		//! Get number of dimensions.
		int GetNumDims (
			) const { return (m_dims.GetNumItems()); }

		//! Get array value type.
		HDF4_VALTYPE GetType (
			) const { return (m_valtype); }

		//! Determine if has 'null' (fill) value set.
		bool HasNullValue (
			) const { return (m_HasNull); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<INT32> m_dims;
		HDF4_VALTYPE m_valtype;
		bool m_HasNull;
		double m_NullValue;
		#endif // GENERATING_DOXYGEN_OUTPUT

		friend class HDF4_ARRAY;
	};


class HDF4_RASTER : public HDF4_OBJECT {
	public:

		class INFO;

		enum INTERLACE {
			INTERLACE_Pixel = 0,
			INTERLACE_Line = 1,
			INTERLACE_Component = 2
			};

		//! Default constructor.
		HDF4_RASTER (
			);

		//! Destructor.
		virtual ~HDF4_RASTER (
			);

		//! Close raster.
		void Close (
			);

		//! Find attribute index by name.
		//! @return Attribute index, -1 if not found, < -1 if error.
		virtual int FindAttrib (
			const char *name						//!< Attribute name to search for
			) const;

		//! Get specific attribute.
		virtual ERRVALUE GetAttrib (
			int index,								//!< Attribute index within raster
			HDF4_ATTRIB& attrib					//!< Attribute instance returned
			) const;

		//! Get raster information.
		ERRVALUE GetInfo (
			INFO& info								//!< Information container returned.
			) const;

		//! Get object name.
		virtual ERRVALUE GetName (
			HDF4_NAME& name						//!< Name returned
			) const;

		//! Get number of attributes in this object.
		virtual int GetNumAttribs (
			) const;

		//! Open raster.
		ERRVALUE Open (
			const HDF4_FILE& file,
			int index
			);

		//! Read from raster.
		//! Note, this will read all components, so allow for that in buffer.
		ERRVALUE Read (
			INT32 ystart,							//!< Starting Y (line) coordinate
			INT32 xstart,							//!< Starting X (column) coordinate
			INT32 ycount,							//!< Number of Y pixels (lines) to read
			INT32 xcount,							//!< Number of X pixels (columns) to read
			void *buffer							//!< Buffer to read into
			) const;

		//! Set interlace mode for reading.
		ERRVALUE SetInterlace (
			INTERLACE interlace
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_rasterid;
		int m_valuesize;

		HDF4_RASTER (const HDF4_RASTER&);
		HDF4_RASTER& operator= (const HDF4_RASTER&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_RASTER::INFO {
	public:

		//! Get number of raster columns;
		INT32 GetNumColumns (
			) const { return (m_NumColumns); }

		//! Get number of raster components.
		INT32 GetNumComponents (
			) const { return (m_NumComponents); }

		//! Get number of raster lines.
		INT32 GetNumLines (
			) const { return (m_NumLines); }

		//! Get value type.
		HDF4_VALTYPE GetType (
			) const { return (m_valtype); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		INT32 m_NumComponents;
		INT32 m_NumColumns;
		INT32 m_NumLines;
		HDF4_VALTYPE m_valtype;
		#endif // GENERATING_DOXYGEN_OUTPUT

		friend class HDF4_RASTER;
	};


class HDF4_TABLE : public HDF4_OBJECT {
	public:

		//! Field information.
		class FIELDINFO {
			public:

				//! Get field name.
				const HDF4_NAME& GetName (
					) const { return (m_name); }

				//! Get offset to start of field in record buffer.
				int GetOffset (
					) const { return (m_offset); }

				//! Get field size.
				int GetSize (
					) const { return (m_size); }

				//! Get field type.
				HDF4_VALTYPE GetType (
					) const { return (m_type); }

				//! Determine if field is a string.
				bool IsString (
					) const { return (m_type == HDF4_VALTYPE_UCHAR8 || m_type == HDF4_VALTYPE_CHAR8); }

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				HDF4_NAME m_name;
				HDF4_VALTYPE m_type;
				int m_size;
				int m_offset;
				#endif // GENERATING_DOXYGEN_OUTPUT

				friend class HDF4_TABLE;
			};

		//! Default constructor.
		HDF4_TABLE (
			);

		//! Destructor.
		virtual ~HDF4_TABLE (
			);

		//! Close table.
		void Close (
			);

		//! Find attribute index by name.
		//! @return Attribute index, -1 if not found, < -1 if error.
		virtual int FindAttrib (
			const char *name						//!< Attribute name to search for
			) const;

		//! Get specific attribute.
		virtual ERRVALUE GetAttrib (
			int index,								//!< Attribute index within table
			HDF4_ATTRIB& attrib					//!< Attribute instance returned
			) const;

		//! Get 'class' name.
		ERRVALUE GetClass (
			HDF4_NAME& classname					//!< Class name returned
			) const;

		//! Get field information array.
		const SIMPLE_ARRAY<FIELDINFO>& GetFieldInfo (
			) const { return (m_FieldInfo); }

		//! Get information about specific field.
		const FIELDINFO& GetFieldInfo (
			int fieldnum							//!< Field number
			) const { return (m_FieldInfo[fieldnum]); }

		//! Get object name.
		virtual ERRVALUE GetName (
			HDF4_NAME& name						//!< Name returned
			) const;

		//! Get number of attributes in this object.
		virtual int GetNumAttribs (
			) const;

		//! Get number of fields in table.
		int GetNumFields (
			) const { return (m_FieldInfo.GetNumItems()); }

		//! Get number of records in table.
		int GetNumRecords (
			) const { return (m_NumRecords); }

		//! Open table.
		ERRVALUE Open (
			const HDF4_FILE& file,				//!< Open HDF4 file
			int index								//!< Table index (reference number)
			);

		//! Read next record from table.
		ERRVALUE Read (
			HDF4_RECORD& record					//!< Record to read into
			) const;

		//! Seek to specified record
		ERRVALUE Seek (
			int recordnum							//!< Record number to seek to
			) const;

	private:
	   #ifndef GENERATING_DOXYGEN_OUTPUT
		int m_tableid;
		int m_NumRecords;
		int m_RecSize;
		SIMPLE_ARRAY<FIELDINFO> m_FieldInfo;

		HDF4_TABLE (const HDF4_TABLE&);
		HDF4_TABLE& operator= (const HDF4_TABLE&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


class HDF4_RECORD {
	public:

		//! Default constructor.
		HDF4_RECORD (
			);

		//! Get string field.
		//! @return Pointer to string field or NULL if not string
		const char *GetString (
			INT32 FieldNum							//!< Field number
			) const;

		//! Get value for numeric field.
		//! @return Double value converted from the field specified
		double GetValue (
			INT32 FieldNum							//!< Field number
			) const;
			
	private:
	   #ifndef GENERATING_DOXYGEN_OUTPUT
		const HDF4_TABLE *m_pTable;
		SIMPLE_ARRAY<UINT8> m_buffer;
		#endif // GENERATING_DOXYGEN_OUTPUT

		friend class HDF4_TABLE;
	};



#endif	//!< INC_MI32_HDF4_H
