/**
 * \file mi32/micache.h
 * \brief MICACHE class definition.
 *
 * \if NODOC
 * $Id: micache.h_v 1.16 2011/10/19 18:01:23 mju Exp $
 *
 * $Log: micache.h_v $
 * Revision 1.16  2011/10/19 18:01:23  mju
 * Remove redundant hdr.
 *
 * Revision 1.15  2011/09/06 21:16:01  mju
 * Exclude macro from dox.
 *
 * Revision 1.14  2007/10/08 23:07:22  scowan
 * Added thread issues documentation.
 *
 * Revision 1.13  2007/08/24 17:51:53  scowan
 * Updated docs for previous changes.
 *
 * Revision 1.12  2007/08/24 17:45:09  scowan
 * Cache code is now thread safe.
 *
 * Revision 1.10  2006/09/07 17:03:04  scowan
 * Added another select method.
 *
 * Revision 1.9  2006/09/07 16:14:25  scowan
 * Added user data pt4r to select, read and write methods.
 *
 * Revision 1.5  2001/05/16 22:04:27  mju
 * Can't have default implementaton for MICACHIO.CacheSave() as DLL export linkage indeterminant.
 *
 * Revision 1.1  2001/05/15 16:56:48  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_MICACHE_H
#define	INC_MI32_MICACHE_H

#ifndef	INC_MI32_CONTAINERITEMSYNC_H
#include <mi32/containeritemsync.h>
#endif

#ifdef MISYSTEMDLL
	#define CLASSEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSEXPORT MI_DLLCLASSIMPORT
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
class MICACHEPRIV;
class MICACHEITEMPRIV;
class MICACHEIO;
#endif // GENERATING_DOXYGEN_OUTPUT


//! Cache manager.
//! This class is exported from a DLL.  Do not attempt to inherit from it.
//! This class is thread safe if the Create() method is passed 'true' for the bSynchronizeAccess parameter.
//! Thread safety is enhanced by use of exclusive locks based on the itemID passed to the Select() call.
//! This means that only one thread at a time can access the cache item given its itemID value.  Therefore 
//! it may not be necessary to use synchronization objects above the MICACHE interface.  However the MICACHEIO
//! virtual methods should use such synchronization objects as necessary, even though the itemID will be locked
//! during the MICACHEIO::CacheLoad() and MICACHEIO::CacheSave() calls.  To avoid deadlock, do not use the same
//! synchronization objects ahead of MICACHE and in MICACHEIO.
class CLASSEXPORT MICACHE {
	public:

		//! Flags for Select() method.
		enum SELECTFLAGS {
			SELECTFLAG_None =			0x0000,
			SELECTFLAG_NewItem =		0x0001,		//!< Item is new, add to cache without checking for existence or reading
			SELECTFLAG_NoRead =		0x0002		//!< Don't read item (use if plan to overwrite entirely)
			};

		//! Information about cache item passed to ReadFunc/WriteFunc and returned by Select().
		class CLASSEXPORT ITEMINFO {
			public:
				//! Default constructor.
				ITEMINFO (
					);

				//! Destructor, releases cache entry if Select() was used.
				~ITEMINFO (
					);

				//! Get cache item ID.
				INT32 GetItemID (
					) const;

				//! Get memory pointer (const).
				const void* GetMemPtr (
					) const;

				//! Get memory pointer (non-const).
				void* GetMemPtr (
					);

				//! Get user data pointer (non-const).
				void* GetUserDataPtr (
					);

				//! Set memory pointer for programmer-allocated cache items.
				void SetMemPtr (
					void *memptr,
					INT32 size
					);

				//! Set item as 'modified' so will be written when required.
				void SetModified (
					);

				//! Set user data pointer.
				void SetUserDataPtr (
					void* UserDataPtr
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				MICACHEITEMPRIV* m_pPriv;
				MITHREAD::CONTAINERITEMSYNC::OBJECT m_ItemSync;
				
				void ReleaseLock ();

				ITEMINFO (const ITEMINFO&);
				ITEMINFO& operator= (const ITEMINFO&);

				friend class MICACHEPRIV;
				friend class MICACHE;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Default constructor.
		//! This performs initialization only.  Use the appropriate Create() method to actually
		//! create the cache.
		MICACHE ();

		//! Destructor.
		~MICACHE ();

		//! Create cache with equal-sized entries.
		//! Allocation will be handled by the MICACHE class.  If items are only to be read
		//! and not modified then the CACHEIO.CacheSave() method does not need to do anything but return 0.
		//! 'NumCache' needs to be at least as much as the number of threads that will use the cache, 
		//! otherwise an ECacheItemsExhausted error may result.
		ERRVALUE Create (
			INT32 NumCache,							//!< Number of items to retain in cache.
			INT32 ItemSize,							//!< Size of each cache item in bytes
			MICACHEIO& ioref,
			INT32 NumHash = 0,						//!< Hash table size, 0 for default
			bool bSynchronizeAccess = false		//!< Set to 'true' if multiple threads will use this cache
			);

		//! Create cache with entry allocation handled by programmer.
		//! The CACHEIO.CacheLoad() method is responsible for allocating the memory for the item
		//! and must use ITEMINFO::SetMemPtr() to set the memory pointer.  The CACHEIO.CacheSave()
		//! method must be implemented and is responsible for freeing the memory for the item.
		//! 'NumCache' needs to be at least as much as the number of threads that will use the cache, 
		//! otherwise an ECacheItemsExhausted error may result.
		ERRVALUE Create (
			INT32 NumCache,						//!< Number of items to retain in cache
			MICACHEIO& ioref,
			INT32 NumHash = 0,						//!< Hash table size, 0 for default
			bool bSynchronizeAccess = false		//!< Set to 'true' if multiple threads will use this cache
			);

		//! Flush all modified entries to external storage.
		//! This may cause the WriteFunc callback to be invoked.
		ERRVALUE Flush (
			);

		//! Flush specified item from cache.  
		//! The item may be flushed depending on the outstanding ITEMINFO instances that reference the item are
		//! still in scope.  If no ITEMINFO instances referencing the item are in scope, the item will be flushed,
		//! otherwise the item will be flushed when the last ITEMINFO that references the item goes out of scope.
		ERRVALUE FlushItem (
			INT32 ItemID
			);

		//! Free the cache.
		void Free (
			);

		//! Determine if cache has been created.
		bool IsCreated (
			) const;

		//! Read data from cache item.
		//! An EInvalidCacheResize error will be returned if this method is called while 
		//! another thread is calling Resize().
		ERRVALUE Read (
			INT32 ItemID,							//!< Item number to read
			INT32 NumBytes,						//!< Number of bytes to read
			void *buffer,							//!< Buffer to read into
			void* UserDataPtr = 0				//!< User data pointer to attach to the cache item before reading
			);

		//! Read data from cache item.
		//! An EInvalidCacheResize error will be returned if this method is called while 
		//! another thread is calling Resize().
		ERRVALUE Read (
			INT32 ItemID,							//!< Item number to read
			INT32 Offset,							//!< Offset from start of cache item
			INT32 NumBytes,						//!< Number of bytes to read
			void *buffer,							//!< Buffer to read into
			void* UserDataPtr = 0				//!< User data pointer to attach to the cache item before reading
			);

		//! Resize the cache.
		//! If the number of items is less than the current number of items then
		//! the WriteFunc may be invoked.
		//! An EInvalidCacheResize error will be returned if this method is called while 
		//! there are outstanding ITEMINFO instances in this or other threads.
		ERRVALUE Resize (
			INT32 NumCache,
			INT32 NumHash = 0
			);

		//! Select cache entry for direct access.
		//! Items selected for direct access are "locked" into memory until released.
		//! For this reason an application should not select more items than necessary
		//! and should not retain the selection outside of the function in which it is
		//! done.  The selection is automatically released when the ITEMINFO is destroyed
		//! or goes out of scope.
		//! An EInvalidCacheResize error will be returned if this method is called while the 
		//! Resize() method is called from another thread.  An ECacheItemsExhausted error will
		//! be returned if the number of outstanding ITEMINFO's exceed the number of cache items.
		ERRVALUE Select (
			INT32 ItemID,
			ITEMINFO& iteminfo,
			SELECTFLAGS flags = SELECTFLAG_None,
			void* UserDataPtr = 0				//!< User data pointer to attach to the cache item before reading
			);

		//! Write data to cache item.
		//! An EInvalidCacheResize error will be returned if this method is called while 
		//! another thread is calling Resize().
		ERRVALUE Write (
			INT32 ItemID,							//!< Item number to write
			INT32 NumBytes,						//!< Number of bytes to write
			const void *buffer,					//!< Buffer to write from
			void* UserDataPtr = 0				//!< User data pointer to attach to the cache item before writing
			);

		//! Write data to cache item.
		//! An EInvalidCacheResize error will be returned if this method is called while 
		//! another thread is calling Resize().
		ERRVALUE Write (
			INT32 ItemID,							//!< Item number to write
			INT32 Offset,							//!< Offset from start of cache item
			INT32 NumBytes,						//!< Number of bytes to write
			const void *buffer,					//!< Buffer to write from
			void* UserDataPtr = 0				//!< User data pointer to attach to the cache item before writing
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		//! This must be only member so can be safely extended in DLL.
		MICACHEPRIV *m_priv;

		//! Unimplemented.
		MICACHE (const MICACHE&);
		MICACHE& operator= (const MICACHE&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//! Interface class for providing access to external storage.
class MICACHEIO {
	protected:
		//! Load cache entry from external storage.
		//! If program is handling allocation this must allocate memory for item and set using MICACHE::ITEMINFO::SetMemPtr().
		virtual ERRVALUE CacheLoad (
			MICACHE::ITEMINFO& iteminfo
			) = 0;

		//! Save cache entry to external storage.
		//! If program is handling allocation this must free any allocated memory used by item.
		virtual ERRVALUE CacheSave (
			MICACHE::ITEMINFO& iteminfo
			) = 0;
	friend class MICACHEPRIV;
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MICACHE::SELECTFLAGS)
#endif

#undef CLASSEXPORT

#endif	// INC_MI32_MICACHE_H
