/**
 * \file mi32/pttriang.h
 * \brief PTTRIANGULATOR definitions
 *
 * \if NODOC
 * $Id: pttriang.h_v 1.26 2012/07/20 21:50:59 mju Exp $
 *
 * $Log: pttriang.h_v $
 * Revision 1.26  2012/07/20 21:50:59  mju
 * Remove unused members.
 *
 * Revision 1.25  2012/07/20 15:01:46  mju
 * Remove unused methods.
 * Mark unused members for deletion.
 *
 * Revision 1.24  2011/11/04 13:15:53  mju
 * Remove redundant hdrs.
 *
 * Revision 1.23  2011/09/06 21:30:22  mju
 * Rename define_enum macro.
 *
 * Revision 1.22  2010/09/29 22:12:35  mju
 * Make all get/is methods const.
 *
 * Revision 1.21  2008/08/20 21:09:34  mju
 * Clarify method names and add computeWorkNodeOfIndex.
 *
 * Revision 1.16  2008/05/12 16:22:04  vdronov
 * added AddPoints
 *
 * Revision 1.15  2008/05/07 22:49:42  vdronov
 * added RULE_NoRestrictions
 *
 * Revision 1.14  2007/06/27 14:47:07  mju
 * Add computeExtents and extents member.
 *
 * Revision 1.13  2006/10/16 19:35:27  vdronov
 * fixed indicies computing
 *
 * Revision 1.12  2006/04/05 17:37:05  vdronov
 * added m_WorkIndices and some other buffers
 *
 * Revision 1.9  2006/01/25 14:42:38  mju
 * Don't pragma warning 4251 as disabled in stddefns.
 *
 * Revision 1.7  2004/05/19 21:33:27  vdronov
 * handle boundaries
 *
 * Revision 1.1  2004/02/27 01:35:11  vdronov
 * Initial revision
 *
 * \endif
*/

#ifndef	INC_MI32_PTTRIANG_H
#define	INC_MI32_PTTRIANG_H

#ifndef	INC_MI32_REGION2D_H
#include <mi32/region2d.h>
#endif

#ifndef	INC_MI32_BITSET_H
#include <mi32/bitset.h>
#endif

#ifndef	INC_MI32_RVCTIN_H
#include <mi32/rvctin.h>
#endif

#ifndef	INC_MI32_QTREE_H
#include <mi32/qtree.h> 
#endif

#include <set>
#include <map>
#include <vector>

#ifdef GEOMDLL
	#define GEOMLIBEXPORT MI_DLLEXPORT
	#define GEOMLIBCLASSEXPORT MI_DLLCLASSEXPORT
#else
	#define GEOMLIBEXPORT MI_DLLIMPORT
	#define GEOMLIBCLASSEXPORT MI_DLLCLASSIMPORT
#endif


//! PTTRIANGULATOR class provides algorithm of triangulation of points with respect of hard edges and clip regions.
class GEOMLIBCLASSEXPORT PTTRIANGULATOR {

	public:

		//! Rule is used to swap diagonal in quadritlateral
		enum RULE {
			RULE_Delaunay = 0,
			RULE_ShortestEdge = 1,
			RULE_NoRestrictions = 2 
			};

		typedef SIMPLE_ARRAY<INT32> BOUNDARY;
		typedef MILIST<BOUNDARY> BOUNDARYLIST;

		#ifndef GENERATING_DOXYGEN_OUTPUT
		enum FLAGS {
			FLAG_ElementBit =		0x01,
			FLAG_EdgeBit =			0x02,
			FLAG_DiagonalBit =	0x04,
			FLAG_Point =			0x00,
			FLAG_Edge =				0x01,
			FLAG_Clip =				0x00,
			FLAG_Hard =				0x02,
			FLAG_TLtoBR =			0x00,
			FLAG_BLtoTR =			0x04
			};
		#endif // GENERATING_DOXYGEN_OUTPUT

		//! Constructor
		PTTRIANGULATOR ();

		//! Destructor
		~PTTRIANGULATOR ();

		//! Assignment operator 
		PTTRIANGULATOR& operator= (				
			const PTTRIANGULATOR& rhs
			);

		//! Add hard edge for triangulation as two DPOINT2Ds
		void AddHardEdge (
			const DPOINT2D& start,
			const DPOINT2D& end
			);

		//! Add hard edge for triangulation as two indices for added before points
		ERRVALUE AddHardEdge (
			const INT32& start, 
			const INT32& end
			);

		//! Add hard line for triangulation as polyline
		void AddHardLine (
			const POLYLINE& polyline
			);

		//! Add point for triangulation
		//! Return index of point added for triangulation
		INT32 AddPoint (
			const DPOINT2D& point
			);

		//! Add points for triangulation
		//! Return index of first point added for triangulation
		INT32 AddPoints (
			const DOUBLE_ARRAY<DPOINT2D>& points
			);

		//! Add points for triangulation
		//! Return index of first point added for triangulation
		INT32 AddPoints (
			const DOUBLE_ARRAY<DPOINT3D>& points
			);

		//! Add points for triangulation
		//! Return index of first point added for triangulation
		INT32 AddPoints (
			const POLYLINE& polyline
			);

		//! Clear all input data.
		void ClearAll ();

		void ClearAllSelection (
		   ) { ClearEdgesSelection(); ClearHardEdgesSelection(); ClearTrianglesSelection(); };

		//! Clear clip region
		void ClearClipRegion (
			) { m_ClipRegion.Clear(); };
											
		void ClearEdgesSelection (
		   ) { m_SelectedEdges.ClearAll(); };

		//! Clear all hard edges and lines
		void ClearHardEdges ();

		void ClearHardEdgesSelection (
			) { m_SelectedHardEdges.ClearAll(); };

		//! Clear all points
		void ClearPoints ();

		void ClearTrianglesSelection (
		   ) { m_SelectedTriangles.ClearAll(); };

		//! Find closest triangle to specified point.
		//! @return Triangle index or -1 if none found.
		INT32 FindClosestTriangle (
			const DPOINT2D& point 
			) const;

		//! Find closest node to specified point, or next closest if same as specified.
		INT32 FindDifferentNode (
			const DPOINT2D& point,
			const INT32 node
			) const;

		//! Find closest edge to specified point.
		INT32 FindEdge (
			const DPOINT2D& point 
			) const;

		//! Find closest hard edge to specified point.
		INT32 FindHardEdge (
			const DPOINT2D& point 
			) const;

		//! Find closest node to specified point.
		INT32 FindNode (
			const DPOINT2D& point 
			) const;

		INT32 FindTriangle (
			const DPOINT2D& point 
			) const;

	 	void GetBoundaries (
			BOUNDARYLIST& boundaries 
			) const;

	 	//! Get edge information for given edge
		//! Return true if given edge exists
		bool GetEdgeInfo (
			const INT32 edge, 
			EDGEINFO& edgeinfo
			) const;

		void GetHardEdges (
			SIMPLE_ARRAY<INT32>& hardedges 
			) const;

	 	//! Get index of node assigned when this node was added for triangulation
	  	//! Return index of node or -1 if node is part of break edge of clip region
		INT32 GetIndexOfNode (
			const INT32 node
			) const;

	 	ERRVALUE GetNodeEdges (
			const INT32 node,
			SIMPLE_ARRAY<INT32>& edges
			) const;

	 	ERRVALUE GetNodeNodes (
			const INT32 node,
			SIMPLE_ARRAY<INT32>& nodes
			) const;

	 	INT32 GetNodeOfIndex (
			const INT32 index
			) const;

		//! Get number of edges in triangulation
		//! Return number of edges
		INT32 GetNumEdges (
			) const { return (m_IsTriangulated ? m_WorkEdges.GetNumItems() : 0); }

		//! Get number of nodes in triangulation
		//! Return number of nodes
		INT32 GetNumNodes (
			) const { return (m_IsTriangulated ? m_WorkPoints.GetNumItems() : 0); }

		INT32 GetNumSelectedEdges (
		   ) const { return (m_SelectedEdges.CountValues()); }

		INT32 GetNumSelectedHardEdges (
		   ) const { return (m_SelectedHardEdges.CountValues()); }

	   INT32 GetNumSelectedTriangles (
		   ) const { return (m_SelectedTriangles.CountValues()); }

		//! Get number of triangles triangulation
		//! Return number of triangles
		INT32 GetNumTriangles (
			) const { return (m_IsTriangulated ? m_WorkTriangles.GetNumItems() : 0); }

	 	//! Get 2D point for given node
		//! Return true if given node exists
		bool GetPoint2D (
			const INT32 node, 
			DPOINT2D& point
			) const;

	 	//! Get 3D point for given node
		//! Return true if given node exists
		bool GetPoint3D (
			const INT32 node, 
			DPOINT3D& point
			) const;

		//! Get rule used for triangulation
		//! Return rule
		RULE GetRule (
			) const { return m_Rule; };

		//! Get tolerance used for triangulation
		//! Return tolerance
		double GetTolerance (
			) const { return m_Tolerance; };

		//! Get edges information for given triangle
		//! Return true if given triangle exists
		bool GetTriangleEdges (
			const INT32 triangle, 
			TRIANGLEEDGES& edges
			) const;

	 	//! Get triangle information for given edge
		//! Return true if given triangle exists
		bool GetTriangleInfo (
			const INT32 triangle, 
			TRIANGLEINFO& triangleinfo
			) const;

		//! Get nodes information for given triangle
		//! Return true if given triangle exists
		bool GetTriangleNodes (
			const INT32 triangle, 
			TRIANGLENODES& nodes
			) const;

		//! Get neighbor triangles information for given triangle
		//! Return true if given triangle exists
		bool GetTriangleTriangles (
			const INT32 triangle, 
			TRIANGLETRIANGLES& triangles
			) const;

		bool IsEdgeSelected (
			const INT32 edge 
			) const { return m_SelectedEdges.GetBit(edge); }

		bool IsHardEdge (
			const INT32 start,
			const INT32 end
			) const;

		bool IsHardEdgeSelected (
			const INT32 edge 
			) const { return m_SelectedHardEdges.GetBit(edge); }

		bool IsSegmentIntersectedByHardEdges (
			const INT32 start,
			const INT32 end
			) const;

		bool IsTriangleSelected (
			const INT32 triangle
			) const { return m_SelectedTriangles.GetBit(triangle); }

		//! Is triangulated
		//! Return true if it is
		bool IsTriangulated (
			) { return m_IsTriangulated; };

		//! Remove selected edges.
		//! @return Number of edges removed or error < 0.
		INT32 RemoveSelectedEdges ();

		//! Remove selected hard edges.
		//! @return Number of edges removed or error < 0.
		INT32 RemoveSelectedHardEdges ();

		//! Remove selected triangles.
		//! @return Number of triangles removed or error < 0.
		INT32 RemoveSelectedTriangles ();

		void SelectEdge (
			const INT32 edge 
			);

		void SelectHardEdge (
			const INT32 edge 
			);

		void SelectTriangle (
			const INT32 triangle
			);

	 	void SetBoundaries (
			const BOUNDARYLIST& boundaries 
			);

		//! Set clip region
		void SetClipRegion (
			const REGION2D& region
			) { m_IsTriangulated = false;	m_ClipRegion = region; };

		//! Triangulate given set of input data for given rule and tolerance
		//! Return error code
		ERRVALUE Triangulate (
			const double tolerance = 0.0, 
			const RULE rule = RULE_Delaunay
			);

		void UnSelectEdge (
			const INT32 edge 
			);

		void UnSelectHardEdge (
			const INT32 edge 
			);

		void UnSelectTriangle (
			const INT32 triangle
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		struct BOUNDSTRUCT {
			POLYLINE m_Polyline;
			DPOINT2D m_Point;
			INT32 m_Parent;
			INT32 m_Level;

			BOUNDSTRUCT () : m_Parent(-1), m_Level(-1) {};
			};

		struct QUADRUPLE {
			QUADRUPLE (
				const INT32 n1,
				const INT32 n2,
				const INT32 n3,
				const INT32 n4
				) :
				node1(n1),
				node2(n2),
				node3(n3),
				node4(n4)
				{
				};

			bool operator < (
				const QUADRUPLE& q
				) const {
				if (node1 < q.node1) return true;
				if (node1 > q.node1) return false;
				if (node2 < q.node2) return true;
				if (node2 > q.node2) return false;
				if (node3 < q.node3) return true;
				if (node3 > q.node3) return false;
				if (node4 < q.node4) return true;
				return false; 
				};
			const INT32 node1;
			const INT32 node2;
			const INT32 node3;
			const INT32 node4;
			};

		typedef std::map<double, DPOINT2D> INTERSECTION;
		typedef std::pair<double, DPOINT2D> INTERSECTIONPAIR;

		RULE m_Rule;
		bool m_IsTriangulated;
		double m_Tolerance;
		DOUBLE_ARRAY<DPOINT3D> m_Points;
		POLYLINELIST m_HardLines;
		DRECT2D m_Extents;
		REGION2D m_ClipRegion;
		bool m_AreHardEdgesExist;
		QUADTREE m_ElementQT;
		SIMPLE_ARRAY<FLAGS> m_ElementStatus;
		INT32 m_NumQTElements;
		QUADTREE m_EdgeQT;
		std::set<INT32> m_Index;
		BITSET m_SelectedEdges;
		BITSET m_SelectedTriangles;
		BITSET m_SelectedHardEdges;
		INTERSECTION m_Intersection;
		INTERSECTION m_LeftPoints;
		INTERSECTION m_RightPoints;
		DOUBLE_ARRAY<DPOINT2D> m_WorkPoints;
		DOUBLE_ARRAY<double> m_WorkZ;
		SIMPLE_ARRAY<INT32> m_WorkNodeOfIndex;			// node for index
		SIMPLE_ARRAY<INT32> m_IndexOfWorkNode;			// index for node
		SIMPLE_ARRAY<EDGEINFO> m_WorkEdges;
		SIMPLE_ARRAY<TRIANGLEEDGES> m_WorkTriangles;
		MILIST<INT32> m_WorkHull;		 
		std::set<INT32> m_DirtyEdges;
		mutable std::set<QUADRUPLE> m_Set;		
		mutable DOUBLE_ARRAY<DPOINT3D> m_TempPoints;   
		mutable SIMPLE_ARRAY<INT32> m_TempElements;		
		mutable SIMPLE_ARRAY<INT32> m_TempEdges;		
		mutable SIMPLE_ARRAY<INT32> m_TempArray;		

		static int ComparePoints(void*, void*, void*);
		static int CompareVectors(const void*, const void*);
		static double DistanceSquaredFromPointToSegement (const DPOINT2D& point, const DPOINT2D& start, const DPOINT2D& end);
		static int GetQuarter(const DPOINT3D* v);

		bool AContainsB (std::vector<PTTRIANGULATOR::BOUNDSTRUCT>& bounds, INT32 A, INT32 B);
		ERRVALUE AddQTElement (const DPOINT2D& start, const DPOINT2D& end, const FLAGS flags);
		ERRVALUE AddSegment (const DPOINT2D& start, const DPOINT2D& end, const FLAGS flags);
		void CheckDirtyEdges (const bool update = false);
		ERRVALUE CheckEdgeIsValid(const INT32 edge);
		ERRVALUE CheckTopologyIsValid();
		ERRVALUE CheckTriangleIsValid(const INT32 triangle);
		bool CircleTest (const INT32 bottomnode, const INT32 leftnode, const INT32 topnode, const INT32 rightnode) const; 
		ERRVALUE ClearStructures ();
		ERRVALUE CompressQTElement ();
		void ComputeExtents ();
		void ComputeWorkNodeOfIndex ();
		ERRVALUE ConstructAllTriangles ();
		double ConvexTest (const INT32 n1, const INT32 n2, const INT32 n3) const;
		bool FillGapEdge (const INT32 edge);
		void FillGapTriangle (const INT32 triangle);
		INT32 FindClosestEdge (const DPOINT2D& point) const;
		ERRVALUE FindFirstTriangle ();
		ERRVALUE FindPrivateEdge (const DPOINT2D& start, const DPOINT2D& end) const;
		ERRVALUE FindPrivateNode (const DPOINT2D& p) const;
		ERRVALUE FindQTElementForEdge (const INT32 edge);
		INT32 GetLeftBottomEdge (const INT32 edge) const;
		INT32 GetLeftNode (const INT32 edge) const;
		INT32 GetLeftTopEdge (const INT32 edge) const;
		INT32 GetRightBottomEdge (const INT32 edge) const;
		INT32 GetRightNode (const INT32 edge) const;
		INT32 GetRightTopEdge (const INT32 edge) const;
		ERRVALUE InsertHardEdge(const INT32 start, const INT32 end);
		ERRVALUE InsertHardEdges ();
		INT32 IntersectSegmentBy (const DPOINT2D& start, const DPOINT2D& end, const INT32 index, double t[2], DPOINT2D point[2]);

		bool IsIntersectedByEdge (
			const DPOINT2D& start, 
			const DPOINT2D& end, 
			const INT32 edge,	
			double& t, 
			DPOINT2D& left,
			DPOINT2D& right
			);

		bool IsPointInside(const std::vector<PTTRIANGULATOR::BOUNDSTRUCT>& bounds, const INT32 maxlevels, const DPOINT2D& point);
		bool IsSwapNeeded (const INT32 edge) const;
		ERRVALUE MakeTriangle (const DPOINT2D& p1, const DPOINT2D& p2, const DPOINT2D& p3);
		ERRVALUE PrepareClipRegion ();
		ERRVALUE PrepareHardLines ();
		ERRVALUE PreparePoints ();
		ERRVALUE RemoveOutsideEdges ();
		void RemovePrivateEdge (const INT32 edge);
		void RemovePrivateTriangle (const INT32 triangle);
		ERRVALUE RemoveQTElement (const INT32 index);
		void SetSelection();
		ERRVALUE SortWorkNodes ();
		ERRVALUE SplitSegementByNodes (const INT32 start, const INT32 end, SIMPLE_ARRAY<INT32>& array);
		void SwapEdge (const INT32 edge, const bool update);
		ERRVALUE TriangulatePolygon (INTERSECTION& Points);
		ERRVALUE UpdateQTEdge ();
		ERRVALUE UpdateQTElement (const INT32 index, const DPOINT2D& start, const DPOINT2D& end, const FLAGS flags);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(PTTRIANGULATOR::FLAGS)
#endif

#endif	// INC_MI32_PTTRIANG_H
