/******************************************************************************
 *
 * \file sml/classmember.h
 * \brief SML CLASSMEMBER
 *
 * \if NODOC
 * $Log: classmember.h_v $
 * Revision 1.5  2011/10/12 14:37:54  mju
 * Remove redundant hdr.
 *
 * Revision 1.4  2009/11/25 18:52:25  mju
 * Doxygen update.
 *
 * Revision 1.3  2007/05/08 15:45:32  dwilliss
 * SML namespace
 *
 * Revision 1.2  2007/05/04 22:34:14  dwilliss
 * moved more stuff here
 *
 * Revision 1.1  2007/05/04 22:30:31  dwilliss
 * Initial revision
 *
 * \endif
*/
/*****************************************************************************/

#ifndef INC_SML_CLASSMEMBER_H
#define INC_SML_CLASSMEMBER_H

#ifndef	INC_MI32_STDDEFNS_H
#include <mi32/stddefns.h>
#endif

#ifndef	INC_SML_STYPE_H
#include <sml/stype.h>
#endif

namespace SML {

union ARG;
class FLAG;
class FUNC;
class CLASS;
class CONTEXT;

#ifdef RVCSYSDLL
	#define RVCSYSLIBEXPORT MI_DLLEXPORT
	#define RVCSYSLIBCLASSEXPORT MI_DLLCLASSEXPORT
#else
	#define RVCSYSLIBEXPORT MI_DLLIMPORT
	#define RVCSYSLIBCLASSEXPORT MI_DLLCLASSIMPORT
#endif

class CLASSMEMBER {
	public:

	//! These are derived from INITYPE constants for convenience, but
	//! wouldn't have to be.  However, the values must stay the same.
	enum TYPE {
		TYPE_ByteBoolean = 1,
		TYPE_IntBoolean	 = 2,
		TYPE_INT8 = 3,
		TYPE_UINT8 = 4,
		TYPE_INT16 = 5,
		TYPE_UINT16 = 6,
		TYPE_INT32 = 7,
		TYPE_UINT32 = 8,
		TYPE_Float = 9,
		TYPE_Double = 10,
		TYPE_String = 11,
		TYPE_Unicode = 12,
		TYPE_Bool	 = 19,
		//! Types not derived from INI types should start at 50 and increase
		TYPE_MISTRING = 50,
		TYPE_Enum = 51
		};

	const char *name;
	SMLSTYPE type;				//!< STYPE
	void (*ReadFunc)(CONTEXT *context, CLASSMEMBER *, void *obj, ARG *ret);	//! Read the member and push onto stack
	void (*WriteFunc)(CONTEXT *context, CLASSMEMBER *, void *obj, ARG *data);	//! Take top of stack and write into 
	FUNC *func;							//!< If it's a function, this points to an SML::FUNC
	const char *ClassName;				//!< Name of member class.  Used to fill in MemberClass later
	UINT32 offset;							//!< Offset into data where member can be found (for simple stuff)
	int subtype;							//!< SML::CLASSMEMBER::TYPE_xxx
	int size;								//!< sizeof(str) for string/Unicode strings
	const char *ResourceType;			//!< eg: XtRDimeision
	const char *ResourceName;			//!< eg: XmNwidth
	FLAG *smlflags;			//!< For resources like XmHORIZONTAL
	const char *description;			//!< Default description which is used to generate the documentation
	CLASS *MemberClass;		//!< The class of the member if it's a class object. Filled in using ClassName
	CLASS *ThisClass;		//!< Pointer to the class this thing is a member of
	};



/*
**	These functions can be used by class members to access simple stuff
*/

//////////////////////////////////////////////////////////////////////////////!
//! 	This function is meant to be used as the ReadFunc of an SML::CLASSMEMBER.
//! 	It assumes the following based on the member's type:
//!		<ul>
//! 	<li><b><tt>STYPE_Num</tt></b>:
//! 			The subtype is one of the SML::CLASSMEMBER::TYPE enums
//! 			for a numeric data type.  The "offset" the offsetof()
//! 			the member.
//! 	<li><b><tt>STYPE_String</tt></b>:
//! 			If the "smlflags" is set, then the "subtype" is assumed
//! 			to be one of the integer SML::CLASSMEMBER::TYPEs and the member
//! 			at "offset" is an integer of the specified type.  The flags
//! 			array is scanned for a flag of the current value and if
//! 			found, the name of the flag is returned to the SML script
//! 			as a string.  If the member's value is not found in the
//! 			smlflags array, then the script is given the string "".
//!
//! 			If the "smlflags" is not set, then subtype must be either
//! 			SML::CLASSMEMBER::TYPE_Unicode, SML::CLASSMEMBER::TYPE_String
//! 			SML::CLASSMEMBER::TYPE_MISTRING.  For TYPE_String and
//! 			TYPE_Unicode, if the "size" member is set, it is assumed that
//! 			the string is defined inline in
//! 			the structure as an arry of that many BYTES (just sizeof())
//! 			If the size is not set, then it is assumed that the offset
//! 			points to a (char*) or (MIUNICODE*).  Note that
//! 			SML::CLASSWriteSimple() will free the pointer before and
//! 			assign a new copy of the string being written.
//! 			The "size" member is ignored for TYPE_MISTRING
//!
//! 	<li><b><tt>STYPE_Enum</tt></b>:
//! 			Behaves like STYPE_String with the smlflags field filled
//! 			in, but instead of getting the list of possible values
//! 			from the smlflags member, it uses "ResourceName" member
//! 			to look up a list of possible values.
//!
//! 	<li><b><tt>STYPE_Class</tt></b>
//! 			Returns a copy to the class data.  SML must know how to free
//!			the class.
//!
//! 	<li><b><tt>STYPE_ClassRef</tt></b>
//! 			Returns a reference to the actual class data (not a copy)
//! 			If "subtype" is 0, then "offset" is assumed to be an offset
//! 			to a POINTER to the thing in question.  If "subtype" is
//! 			non-zero, then offset is assumed to be an offset to a
//! 			structure within the structure.
//! </ul>
//!
RVCSYSLIBEXPORT void CLASSReadSimple (
	CONTEXT *,
	CLASSMEMBER *,
	void *,
	ARG *
	);

//////////////////////////////////////////////////////////////////////////////!
//! 	This function is meant to be used as the WriteFunc of an SML::CLASSMEMBER.
//! 	Uses the same assumptions as SML::CLASSReadSimple(), with the following
//! 	exceptions:
//!
//! 	STYPE_Class.
//! 			If subtype is non-zero, the offset is assumed to point to
//! 			a structure within a structure.  In this case, the assignment
//! 			will be done by memcpy().  To know how much to memcpy, there
//! 			are two options.
//! 				1. you can set the "size" for the member to the sizeof().
//! 				2. If the member is of another class defined within SML
//! 				   (ie, the "ClassName" is set), it will use that class's
//! 				   size
//! 			If the subtype is 0, the offset is an offset to a POINTER
//! 			which will simply be assigned to whatever pointer it's given.
//!
RVCSYSLIBEXPORT void CLASSWriteSimple (
	CONTEXT *,
	CLASSMEMBER *,
	void *,
	ARG *
	);

//////////////////////////////////////////////////////////////////////////////!
//! 	This function is meant to be used as the WriteFunc of an SML::CLASSMEMBER
//!		if you don't want anything to write to it.  This is only necessary
//!		for STYPE_Class members where the readfunc is returning a pointer.
//!		For example:  If class FRED had a read-only member "name", to prevent
//!		@code
//!			fred.name = "george"
//!		@endcode
//!		you only have to set the member's WriteFunc to NULL.  But if fred
//!		also has a readonly member EyeColor, to prevent
//!		@code
//!			fred.EyeColor.Name = "red"
//!		@endcode
//!		Setting the WriteFunc to NULL isn't enough.  The above only reads
//!		EyeColor and leaves a pointer to it on the stack.  The write in this
//!		case is on a class COLOR.  But if you set EyeColor's WriteFunc to
//!		SML::CLASSReadOnly, the SML parser will prevent the above assignment.
//!
//!
RVCSYSLIBEXPORT void CLASSReadOnly (
	CONTEXT *,
	CLASSMEMBER *,
	void *,
	ARG *
	);

//! Function to use as SML::CLASSMEMBER.ReadFunc to simply return pointer to class/struct itself.
//! This allows a portion or member of a class/struct to be accessed as a member while still
//! retaining the pointer to the class/struct itself.
void CLASSReadPassThrough (CONTEXT *, CLASSMEMBER *, void *vdata, ARG *ret);

//////////////////////////////////////////////////////////////////////////////!
//! 	This function is meant to be used as the ReadFunc of an SML::CLASSMEMBER.
//! 	which must be a UINT32 at the "offset" specified in the SML::CLASSMEMBER.
//!		This is used when a ULONG has multiple bit-field flags ORd together
//!		and you want the SML Class to appear to have boolean members for
//!		each flag.  The SML::CLASSMEMBER's "subtype" wil be AND'd with the
//!		UINT32 for the member, and will return 1 (true) or 0 (false).  To
//!		reverse the symantics, set the SML::CLASSMEMBER's size to 1.
//!
//!		This function is NOT compatible with C++ enums because you can't
//!		depend on them being the same as a UINT32
//!
RVCSYSLIBEXPORT void CLASSReadULONGFlag (
	CONTEXT *,
	CLASSMEMBER *,
	void *,
	ARG *
	);

//////////////////////////////////////////////////////////////////////////////!
//! 	This function is meant to be used as the WriteFunc of an SML::CLASSMEMBER.
//! 	It's just like SML::CLASSReadULONGFlag() but it writes.
//!
RVCSYSLIBEXPORT void CLASSWriteULONGFlag (
	CONTEXT *,
	CLASSMEMBER *,
	void *,
	ARG *
	);

//! Function to use as WriteFunc of SML::CLASSMEMBER to update an INT16 table number field
//! from an SML DBTABLEINFO class (represented by SML::FIELDDATA structure).
//!
//! The SML::CLASSMEMBER definition have the following settings:
//! <ul>
//!		<li>type			-	STYPE_Class
//!		<li>ReadFunc		-	NULL
//!		<li>WriteFunc	-	SML::CLASSWriteTableNum
//!		<li>ClassName	-	SML::CLASSNAME_DBTABLEINFO
//!		<li>offset		-	Offset of INT16 table number in associated struct/class
//!	</ul>
void CLASSWriteTableNum (CONTEXT *context, CLASSMEMBER *member, void *vdata, ARG *arg);

//////////////////////////////////////////////////////////////////////////////!


//////////////////////////////////////////////////////////////////////////////!
//////////////////////////////////////////////////////////////////////////////!
//!
//!		WARNING:  Experamental macros
//!
/*****************************************************************************/
//!
//!	Macros for def'n of class members which are directly accessable.
//!	These macros take the following parameters... (not all macros take all parameters)
//!
//!	smlname		:	the name of the member as it will appear in sml (in quotes)
//!	Class			:	C++ class/struct name
//!	member		:	C++ class/struct member name
//!	subtype		:	An SML::CLASSMEMBER::TYPE value
//!	classname	:  For members which are classes themselves, this is the name of
//!						the member's class as it appears in SML (in quotes)
//!	size			:	Only needed for strings.  Size of the string in BYTES
//!	flagval		:	For UINT32 Flags, a constant flag value to check for.
//!	invert		:  For UINT32 Flags. If 1, the SML meaning of the flag is opposite
//!						the meaning implied by flagval
//!	smlflags		:	For SML::FLAG: SML::FLAG array of flags
//!	enumname		:	For ENUM: name (in quotes) of ENUM list to lookup eg: "UNITS::Distance"
//!						Enums have the advantage that the can be defined in a central location and
//!						the list can be generated as-needed from some other source (like units.txt)
//!	mask			:	For SML::FLAG or ENUM: Optional bitmask to allow a subset of bits
//!	desc			:	Default description.  This is used to build smlclass.txt
//!
//! Macros for easy def'n of class members which are directly accessable
//!
//!	The macros exist in 2 flavors for each type of data:
//!		SMLCLASSMEMBER_XXXXX_RO		: Read Only
//!		SMLCLASSMEMBER_XXXXX_RW		: Read/Write
//!
//!	It is also possible to have class members which can only be written, but
//!	it requires more than the simple macro defn's here.  For an example of
//!	such a class, see the COLOR class in qqstruct.c
//!
//!	The available datatypes are...
//!		SMLCLASSMEMBER_NUM_XX
//!		SMLCLASSMEMBER_STR_PTR_XX
//!		SMLCLASSMEMBER_STR_FIXEDSIZE_XX
//!		SMLCLASSMEMBER_STRUCT_PTR_XX
//!		SMLCLASSMEMBER_STRUCT_INLINE_XX
//!		SMLCLASSMEMBER_UINT32FLAG_XX
//!		SMLCLASSMEMBER_SMLFLAG_XX
//!		SMLCLASSMEMBER_ENUM_XX
//!
#define SMLCLASSMEMBER_NUM_RO(smlname, Class, membername, subtype, desc) \
	{ smlname, STYPE_Num, SMLCLASSReadSimple, NULL, NULL, NULL, offsetof(Class, membername), subtype, 0, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_NUM_RW(smlname, Class, membername, subtype, desc) \
	{ smlname, STYPE_Num, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, NULL, offsetof(Class, membername), subtype, 0, NULL, NULL, NULL, desc}

//! String classes where the member is a (char*) or (MIUNICODE*)
#define SMLCLASSMEMBER_STR_PTR_RO(smlname, Class, membername, subtype, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, NULL, NULL, NULL, offsetof(Class, membername), subtype, 0, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_STR_PTR_RW(smlname, Class, membername, subtype, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, NULL, offsetof(Class, membername), subtype, 0, NULL, NULL, NULL, desc}

//! String classes where the member is of a fixed size as part of the struct/class
#define SMLCLASSMEMBER_STR_FIXEDSIZE_RO(smlname, Class, membername, subtype, sizeOf, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, NULL, NULL, NULL, offsetof(Class, membername), subtype, sizeOf, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_STR_FIXEDSIZE_RW(smlname, Class, membername, subtype, sizeOf, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, NULL, offsetof(Class, membername), subtype, sizeOf, NULL, NULL, NULL, desc}

//! Member is a pointer to a class/struct
#define SMLCLASSMEMBER_STRUCT_PTR_RO(smlname, Class, membername, classname, desc) \
	{ smlname, STYPE_ClassRef, SMLCLASSReadSimple, NULL, NULL, classname, offsetof(Class::membername), 0, 0, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_STRUCT_PTR_RW(smlname, Class, membername, classname, desc) \
	{ smlname, STYPE_ClassRef, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, classname, offsetof(Class::membername), 0, 0, NULL, NULL, NULL, desc}

//! Member is a class, not a pointer to a class
#define SMLCLASSMEMBER_STRUCT_INLINE_RO(smlname, Class, membername, classname, desc) \
	{ smlname, STYPE_ClassRef, SMLCLASSReadSimple, SMLCLASSReadOnly, NULL, classname, offsetof(Class, membername), 1, 0, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_STRUCT_INLINE_RW(smlname, Class, membername, classname, desc) \
	{ smlname, STYPE_ClassRef, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, classname, offsetof(Class, membername), 1, 0, NULL, NULL, NULL, desc}

//! Member which is a flag in an UINT32
#define SMLCLASSMEMBER_UINT32FLAG_RO(smlname, Class, membername, flagval, invert, desc) \
	{ smlname, STYPE_Num, SMLCLASSReadULONGFlag, NULL, NULL, classname, offsetof(Class, membername), flagval, invert, NULL, NULL, NULL, desc}

#define SMLCLASSMEMBER_UINT32FLAG_RW(smlname, Class, membername, flagval, invert, desc) \
	{ smlname, STYPE_Num, SMLCLASSReadULONGFlag, SMLCLASSWriteULONGFlag, NULL, classname, offsetof(Class, membername), flagval, invert, NULL, NULL, NULL, desc}

//! SMLFLAG flags
#define SMLCLASSMEMBER_SMLFLAG_RO(smlname, Class, membername, subtype, smlflags, mask, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, NULL, NULL, NULL, offsetof(Class, membername), subtype, size, NULL, NULL, smlflags, desc}

#define SMLCLASSMEMBER_SMLFLAG_RW(smlname, Class, membername, subtype, smlflags, mask, desc) \
	{ smlname, STYPE_String, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, NULL, offsetof(Class, membername), subtype, mask, NULL, NULL, smlflags, desc}

//! SMLFLAG flags (using named enums)
#define SMLCLASSMEMBER_ENUM_RO(smlname, Class, membername, subtype, enumname, mask, desc) \
	{ smlname, STYPE_Enum, SMLCLASSReadSimple, NULL, NULL, NULL, offsetof(Class, membername), subtype, size, NULL, enumname, NULL, desc}

#define SMLCLASSMEMBER_ENUM_RW(smlname, Class, membername, subtype, enumname, mask, desc) \
	{ smlname, STYPE_Enum, SMLCLASSReadSimple, SMLCLASSWriteSimple, NULL, NULL, offsetof(Class, membername), subtype, mask, NULL, enumname, NULL, desc}

//! The rest of the member struct gets filled in by the installer
//! STYPE_Constant is wrong, but will also be replaced by installer
#define SMLCLASSMETHOD(func) { #func, STYPE_Constant, NULL, NULL, &FUNC_##func }

} // end of SML namespace


typedef SML::CLASSMEMBER SMLCLASSMEMBER;

#define SMLCLASSWriteTableNum SML::CLASSWriteTableNum
#define SMLCLASSReadSimple SML::CLASSReadSimple 
#define SMLCLASSWriteSimple SML::CLASSWriteSimple 
#define SMLCLASSReadOnly SML::CLASSReadOnly 
#define SMLCLASSReadPassThrough SML::CLASSReadPassThrough 
#define SMLCLASSReadULONGFlag SML::CLASSReadULONGFlag 
#define SMLCLASSWriteULONGFlag SML::CLASSWriteULONGFlag


#undef RVCSYSLIBEXPORT
#undef RVCSYSLIBCLASSEXPORT

#endif		//  INC_SML_CLASSMEMBER_H


