/**
 * \file spatmod/imagedimensions.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::DIMENSIONS definition.
 *
 * \if NODOC
 * $Id: imagedimensions.h_v 1.4 2008/11/06 15:47:07 mju Exp $
 *
 * $Log: imagedimensions.h_v $
 * Revision 1.4  2008/11/06 15:47:07  mju
 * Add equality operators.
 *
 * Revision 1.3  2008/08/28 13:54:19  mju
 * Fix doxygen file ref.
 *
 * Revision 1.2  2006/11/02 15:43:58  mju
 * Use spans class.
 *
 * Revision 1.1  2006/06/05 13:17:04  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEDIMENSIONS_H
#define	INC_SPATMOD_IMAGEDIMENSIONS_H

#ifndef	INC_SPATMOD_IMAGESPANS_H
#include <spatmod/imagespans.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Tile position in row or column.
enum TILEPOSITION {
	TILEPOSITION_First,						//!< First tile in row or column
	TILEPOSITION_Interior,					//!< Interior tile
	TILEPOSITION_Last							//!< Last tile in row or column
	};

//=====================================================================================================================

//! Container for defining image dimensions.
class DIMENSIONS {
	public:

		//! Constructor
		DIMENSIONS ();

		//! Get column spans.
		const SPANS& GetColumnSpans (
			) const { return (m_SpansColumn); }

		//! Get number of columns of tiles "across" the image.
		//! @return Number of columns of tiles.
		INT32 GetNumColumnsOfTiles (
			) const { return (m_SpansColumn.GetNumSpans()); }

		//! Get number of columns per tile.
		//! Note that first and last tiles may have different sizes than each other or interior tiles.
		//! If GetNumColumnsOfTiles is less than 3 then will return 0 for TILEPOSITION_Interior.
		//! If GetNumColumnsOfTiles is less than 2 then the "last" tile will have same dimension as "first" tile.
		//! @return Number of columns per tile.
		double GetNumColumnsPerTile (
			TILEPOSITION tileposition				//!< Tile position
			) const;

		//! Get number of rows of tiles "down" the image.
		//! @return Number of rows of tiles.
		INT32 GetNumRowsOfTiles (
			) const { return (m_SpansRow.GetNumSpans()); }

		//! Get number of rows per tile.
		//! Note that first and last tiles may have different sizes than each other or interior tiles.
		//! If GetNumRowsOfTiles is less than 3 then will return 0 for TILEPOSITION_Interior.
		//! If GetNumRowsOfTiles is less than 2 then the "last" tile will have same dimension as "first" tile.
		//! @return Number of rows per tile.
		double GetNumRowsPerTile (
			TILEPOSITION tileposition				//!< Tile position
			) const;

		//! Get row spans.
		const SPANS& GetRowSpans (
			) const { return (m_SpansRow); }

		//! Get total number of columns in image.
		//! @return Total image columns.
		UINT32 GetTotalColumns (
			) const { return (static_cast<UINT32>(m_SpansColumn.GetTotalSize())); }

		//! Get total number of rows in image.
		//! @return Total image rows.
		UINT32 GetTotalRows (
			) const { return (static_cast<UINT32>(m_SpansRow.GetTotalSize())); }

		//! Test if DIMENSIONS are equal.
		bool IsEqual (
			const DIMENSIONS& rhs
			) const { return (m_SpansRow==rhs.m_SpansRow && m_SpansColumn==rhs.m_SpansColumn); }

		//! Set tile dimensions.
		ERRVALUE SetTileDimensions (
			double ColumnsPerTile,
			double RowsPerTile,
			double ColumnsPerTileFirst = 0,
			double RowsPerTileFirst = 0,
			double ColumnsPerTileLast = 0,
			double RowsPerTileLast = 0
			);

		//! Set image dimensions.
		//! Will reset tile dimensions to match image dimensions (resulting in "untiled" image).
		ERRVALUE SetTotalDimensions (
			UINT32 TotalColumns,
			UINT32 TotalRows
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		SPANS m_SpansRow;
		SPANS m_SpansColumn;

		#endif	// GENERATING_DOXYGEN_OUTPUT

	};

inline bool operator== (
	const DIMENSIONS& lhs,
	const DIMENSIONS& rhs
	) { return (lhs.IsEqual(rhs)); }

inline bool operator!= (
	const DIMENSIONS& lhs,
	const DIMENSIONS& rhs
	) { return (!lhs.IsEqual(rhs)); }

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	 // INC_SPATMOD_IMAGEDIMENSIONS_H
