/**
 * \file imagefiltermaskselect.h <spatmod/imagefiltermaskselect.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_MASKSELECT definition.
 *
 * \if NODOC
 * $Id: imagefiltermaskselect.h_v 1.1 2006/08/03 13:50:47 mju Exp $
 *
 * $Log: imagefiltermaskselect.h_v $
 * Revision 1.1  2006/08/03 13:50:47  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERMASKSELECT_H
#define	INC_SPATMOD_IMAGEFILTERMASKSELECT_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Filter to select between images using mask.
//! All source images must have the same dimension.
//! The pixel properties of both non-mask images must match.
//! The mask image must have a single sample per pixel.
//! The georeference from the first source that has any will be used.
//! Invalid mask pixels will result in invalid result pixels, otherwise
//! validity will be transferred from selected source based on mask.
class FILTER_MASKSELECT : public FILTER {
	public:

		//! Construct with same ownership for all source stages.
		FILTER_MASKSELECT (
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			STAGE *pSourceImageZero,				//!< Source image stage selected for zero mask pixels
			STAGE *pSourceImageNonZero,			//!< Source image stage selected for non-zero mask pixels
			bool TakeOwnership						//!< Filter takes ownership of all source stages and will delete when needed
			);

		//! Construct with different ownership for each source stage.
		FILTER_MASKSELECT (
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			STAGE *pSourceImageZero,				//!< Source image stage selected for zero mask pixels
			bool TakeOwnershipImageZero,			//!< Filter takes ownership and will delete when needed
			STAGE *pSourceImageNonZero,			//!< Source image stage selected for non-zero mask pixels
			bool TakeOwnershipImageNonZero		//!< Filter takes ownership and will delete when needed
			);

		// Destructor.
		virtual ~FILTER_MASKSELECT ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		typedef FILTER BASECLASS;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_MASKSELECT (const FILTER_MASKSELECT&);
		FILTER_MASKSELECT& operator= (const FILTER_MASKSELECT&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERMASKSELECT_H

