/**
 * Demonstrate use of TNTsdk object display functions
 *
 * This program allows the user to view one or more spatial data layers.
 *
 * The user can select a text file to read x/y coordinates
 * of symbols to be plotted over the selected objects.  The map projection for
 * these symbols may also be set via the File / Projection option.
 *
 * If the points are to be read from a file generated by a GPS device the
 * function to read the file could be altered to match a specific format and
 * the projection option could be removed if the file format contains this
 * information or it is always a specific projection.
**/

#include <mgui/appmain.h>
#include <mi32/coordtrans.h>
#include <mi32/mg2x.h>
#include <mi32/filetext.h>
#include <mi32/textid.h>
#include "toolskel.h"
#include <gre/viewlegacy.h>
#include <gre/group.h>
#include <gre/system.h>
#include <gre/dlglayermgr.h>
#include <mgui/dlggetobject.h>
#include <memory>

#include <mi32/xdefns.h>
#include <Xm/Form.h>

static int PrivIndex=-1;

#define	SYMBOLRADIUS	4


//=====================================================================================================================
//		APP_OBJVIEW - Begin


class APP_OBJVIEW : public MGUI::APP_BASE {
	public:
		APP_OBJVIEW ();
	private:

		Widget m_toplevel;
		GRE::GROUP *m_group;
		GRE::VIEW_LEGACY *m_view;
		SIMPLE_ARRAY<DPOINT2D> m_SymbolPoint;
		SPATREF::COORDREFSYS m_CoordRefSys;
		FILEPATH m_SymbolFilePath;

		static void CB_MenuFileGetCRS (Widget,void*,void*);
		static void CB_MenuFileReadPoints (Widget,void*,void*);
		static void CB_OpenLayerMgr (Widget,void*,void*);
		void CreateMainWindow ();
		static void GRECB_View (const GRE::MSG *msg, void *vpApp);
		void HandleMsg (const GRE::MSG *msg);

		// Overrides from APP_BASE.
		virtual void v_ExitInstance ();
		virtual const char* v_GetProcessName () const;
		virtual bool v_InitInstance ();
	};


APP_OBJVIEW::APP_OBJVIEW (
	) :
	m_toplevel(0),
	m_group(0),
	m_view(0)
	{
	}


// Callback for File / Projection... menu option
void APP_OBJVIEW::CB_MenuFileGetCRS (
	Widget w,
	void *,
	void *
	) {
	GRE::VIEW_LEGACY *view = static_cast<GRE::VIEW_LEGACY*>(MxGetUserDataPtr(w));
	APP_OBJVIEW* pApp = static_cast<APP_OBJVIEW*>(view->GetPrivPtr(PrivIndex));
	SPATREF::DlgGetCoordRefSys(pApp->m_toplevel,pApp->m_CoordRefSys);
	if (pApp->m_SymbolPoint.GetNumItems() > 0) pApp->m_view->RestoreFromMemory();
	return;
	}


// Callback for File / Projection... menu option
void APP_OBJVIEW::CB_MenuFileReadPoints (
	Widget w,
	void *,
	void *
	) {
	GRE::VIEW_LEGACY *view = static_cast<GRE::VIEW_LEGACY*>(MxGetUserDataPtr(w));
	APP_OBJVIEW* pApp = static_cast<APP_OBJVIEW*>(view->GetPrivPtr(PrivIndex));
	ERRVALUE err;

	if ((err = MGUI::DlgGetFile(pApp->m_toplevel,TEXTID_Prompt_File_Read,"*",pApp->m_SymbolFilePath)) < 0) return;

	FILE_TEXT file;
	STRUTF8 strutf8;
	if ((err = file.Open(pApp->m_SymbolFilePath)) < 0) goto Error;
	pApp->m_SymbolPoint.Clear();

	while ((err = file.ReadLine(strutf8)) >= 0) {
		if (strutf8[0] == '#') continue;		// Used to indicate a comment
		DPOINT2D pt;
		if (sscanf(strutf8,"%lf %lf",&pt.x,&pt.y) == 2) {
			if ((err = pApp->m_SymbolPoint.Append(pt)) < 0) goto Error;
			}
		}
	pApp->m_view->RestoreFromMemory();

Error:
	if (err < -1) MdlgShowErrorCode(NULL,err);
	return;
	}


// Open layer manager dialog
void APP_OBJVIEW::CB_OpenLayerMgr (
	Widget ,
	void *vpApp,
	void *
	) {
	APP_OBJVIEW* pApp = static_cast<APP_OBJVIEW*>(vpApp);
	GRE::DLG_LAYERMGR::OpenSimple(pApp->m_toplevel,pApp->m_group,TEXTID__None);
	return;
	}


void APP_OBJVIEW::CreateMainWindow (
	) {
	static MENUITEM MenuItemFile[] = {
		{ "read_points_", 0, CB_MenuFileReadPoints },
		{ "projection_",  0, CB_MenuFileGetCRS    },
		{ "-" },
		{ "exit",         0, MxcbSetExit },
		{ NULL }
		};
	static MENUBARITEM MenuBar[] = {
		{ "file", MenuItemFile },
		{ "view" },
		{ "tool" },
		{ "options" },
		{ NULL }
		};
	static BUTTONITEM IconItemExtra[] = {
		{ "layer_controls_", 0, CB_OpenLayerMgr, NULL, NULL, NULL, NULL, "layer_controls" },
		{ NULL }
		};

	Arg arg[20];
	int n;

	n = 0;
	XtSetArgI(arg,n,XmNresizePolicy,XmRESIZE_GROW);
	Widget mainform = XmCreateForm(m_toplevel,"mainform",arg,n);

	n = 0;
	m_view = new GRE::VIEW_LEGACY;
	m_view->CreateX(m_group,mainform,arg,n,0,0,"MainView");
	m_view->SetPrivPtr(PrivIndex,this);
	m_view->SetViewFlags(m_view->GetViewFlags() | VIEWFLAG_NoDestroyShell);

	GRE_ToolAddExample(m_view);

	m_view->AddToolIcons(false);

	for (BUTTONITEM *bi = IconItemExtra; bi->label!=NULL; ++bi) {
		bi->callback_data = (void*) this;
		}
	m_view->AddIconRowItems(IconItemExtra);
	m_view->CreateMenuBar(mainform,MenuBar);

	n = 0;
	XtSetArgI(arg,n,XmNminHeight,100);
	XtSetArgI(arg,n,XmNminWidth,100);
	XtSetValues(XtParent(mainform),arg,n);

	XtManageChild(mainform);
	XtRealizeWidget(m_toplevel);

	m_view->CallbackAdd(GRECB_View,this,GRE::OBJECT::TYPE_View);

	return;
	}


// [static] Handle GRE messages.
void APP_OBJVIEW::GRECB_View (
	const GRE::MSG *msg,
	void *vpApp
	) {
	static_cast<APP_OBJVIEW*>(vpApp)->HandleMsg(msg);
	}


// Handle GRE messages.
void APP_OBJVIEW::HandleMsg (
	const GRE::MSG *msg
	) {
	ERRVALUE err = 0;
	COLOR color(COLOR::RED);

	if (m_SymbolPoint.GetNumItems() == 0) return;		// No symbols selected yet
	switch (msg->GetAction()) {
		case GRE::MSG::ACTION_DrawEnd:
		case GRE::MSG::ACTION_Expose:
		case GRE::MSG::ACTION_RestoreEnd: {
			if (m_view->GetDrawingDeviceScreen() == NULL) return;
			std::auto_ptr<MGD::CONTEXT> dc(m_view->GetDrawingDeviceScreen()->CreateContext());
			SPATREF::COORDTRANS TransToGroup;
			if ((err = TransToGroup.SetupAutoCRS(m_CoordRefSys,m_group->GetUsedCoordRefSys())) < 0) break;
			dc->SetColor(color);
			for (UINT32 i = 0; (i < m_SymbolPoint.GetNumItems()); ++i) {
				DPOINT2D dpoint(m_SymbolPoint[i]);
				if ((err = TransToGroup.ConvertForward(&dpoint,1)) < 0) break;
				if ((err = m_view->TransGroupToDisp(m_group,&dpoint,&dpoint,1)) < 0) break;
				INT32 x = FAST_ROUND_LIMITED(dpoint.x);
				INT32 y = FAST_ROUND_LIMITED(dpoint.y);
				dc->FillCircle(x,y,SYMBOLRADIUS);
				}
			break;
			}
		}
	
	if (err < 0) MdlgShowErrorCode(NULL,err);
	return;
	}


void APP_OBJVIEW::v_ExitInstance (
	) {
	m_group->DestroyAllLayers();
	m_view->Destroy();
	return;
	}


const char* APP_OBJVIEW::v_GetProcessName (
	) const {
	return ("objview");
	}


bool APP_OBJVIEW::v_InitInstance (
	) {
	m_toplevel = MxInitLegacy(LEGACYINIT_NoRealize);

	PrivIndex = GRE::SYSTEM::AllocPrivIndex();

	m_group = new GRE::GROUP;
	m_group->SetPrivPtr(PrivIndex,this);

	CreateMainWindow();

	return (true);
	}


//		APP_OBJVIEW - End
//=====================================================================================================================
//		MAIN application instance class

class MAIN : public MGUI::APP_MAIN {
	public:
		MAIN () : m_pApp(new APP_OBJVIEW) {}
		~MAIN () { }
	private:
		APP_OBJVIEW *m_pApp;
	};

static MAIN MainApp;

//=====================================================================================================================
